# TODO:
# - details, edit, delete URLs should be signed
#
#
import base64
import copy
import datetime
import functools
import re
from urllib.parse import urlparse

from pydal.objects import Expression, Field, FieldVirtual
from pydal.querybuilder import QueryBuilder
from yatl.helpers import (
    CAT,
    DIV,
    FORM,
    INPUT,
    OPTION,
    SELECT,
    SPAN,
    TABLE,
    TAG,
    TBODY,
    TD,
    TH,
    THEAD,
    TR,
    XML,
    A,
    I,
)

from py4web import HTTP, URL, redirect, request, safely
from py4web.utils.form import Form, FormStyleDefault, join_classes
from py4web.utils.param import Param

NAV = TAG.nav
HEADER = TAG.header


def title(text):
    """Turns text into a title"""
    return str(text).replace("_", " ").title()


def safe_int(text, default):
    """try parse txt into an int else returns the default"""
    try:
        return int(text)
    except (ValueError, TypeError):
        return default


def query_join(a, b):
    return a + ("&" if "?" in a else "?") + b


def make_default_search_query(table):
    field_aliases = {
        str(field.label).replace(" ", "_").lower(): field.name
        for field in table
        if field.readable
    }
    builder = QueryBuilder(table, field_aliases=field_aliases)
    return ["Query", lambda text, builder=builder: builder.parse(text), None]


def strip_field_type(type_name, regex=re.compile(r"^\w+")):
    """list:string -> list, reference table->reference, decimal(3,7)->decimal"""
    return regex.match(str(type_name)).group()


class GridClassStyle:
    """
    Default grid style
    Internal element names match default class name, other classes can be added
    Style use should be minimized since it cannot be overridden by CSS
    """

    classes = {
        "grid-wrapper": "grid-wrapper",
        "grid-header": "grid-header",
        "grid-new-button": "grid-new-button info",
        "grid-search": "grid-search",
        "grid-table-wrapper": "grid-table-wrapper",
        "grid-table": "grid-table",
        "grid-sorter-icon-up": "grid-sort-icon-up fas fa-sort-up",
        "grid-sorter-icon-down": "grid-sort-icon-down fas fa-sort-down",
        "grid-thead": "",
        "grid-tr": "",
        "grid-th": "",
        "grid-td": "",
        "grid-td-buttons": "",
        "grid-back-button": "info",
        "grid-button": "info",
        "grid-details-button": "grid-details-button info",
        "grid-edit-button": "grid-edit-button info",
        "grid-delete-button": "grid-delete-button info",
        "grid-search-button": "grid-search-button",
        "grid-clear-button": "grid-clear-button",
        "grid-footer": "grid-footer",
        "grid-info": "grid-info",
        "grid-pagination": "grid-pagination",
        "grid-pagination-button": "grid-pagination-button info",
        "grid-pagination-button-current": "grid-pagination-button-current default",
        "grid-cell-type-string": "grid-cell-type-string",
        "grid-cell-type-text": "grid-cell-type-text",
        "grid-cell-type-boolean": "grid-cell-type-boolean",
        "grid-cell-type-float": "grid-cell-type-float",
        "grid-cell-type-decimal": "grid-cell-type-decimal",
        "grid-cell-type-int": "grid-cell-type-int",
        "grid-cell-type-date": "grid-cell-type-date",
        "grid-cell-type-time": "grid-cell-type-time",
        "grid-cell-type-datetime": "grid-cell-type-datetime",
        "grid-cell-type-upload": "grid-cell-type-upload",
        "grid-cell-type-list": "grid-cell-type-list",
        "grid-cell-type-id": "grid-cell-type-id",
        # specific for custom form
        "grid-search-form": "grid-search-form",
        "grid-search-form-table": "grid-search-form-table",
        "grid-search-form-tr": "grid-search-form-tr",
        "grid-search-form-td": "grid-search-form-td",
        "grid-search-form-input": "grid-search-form-input",
        "grid-search-form-select": "grid-search-form-select",
        "grid-search-form-error": "py4web-validation-error",
        "grid-search-boolean": "grid-search-boolean",
        "grid-header-element": "grid-header-element info",
        "grid-footer-element": "grid-footer-element info",
    }

    @classmethod
    def get(cls, element, default=None):
        """returns a dict with _class and _style for the element name"""
        return cls.classes.get(element, element if default is None else default)


class GridClassStyleBulma(GridClassStyle):
    """The Grid style for Bulma"""

    classes = {
        "grid-wrapper": "grid-wrapper field",
        "grid-header": "grid-header pb-2 is-clearfix",
        "grid-new-button": "grid-new-button button",
        "grid-search": "grid-search is-pulled-right pb-2",
        "grid-table-wrapper": "grid-table-wrapper table_wrapper table-container mb-1",
        "grid-table": "grid-table table is-bordered is-striped is-hoverable is-fullwidth is-clearfix",
        "grid-sorter-icon-up": "grid-sort-icon-up fas fa-sort-up is-pulled-right",
        "grid-sorter-icon-down": "grid-sort-icon-down fas fa-sort-down is-pulled-right",
        "grid-thead": "",
        "grid-tr": "",
        "grid-th": "",
        "grid-td": "is-small",
        "grid-td-buttons": "is-small is-narrow",
        "grid-back-button": "grid-button button",
        "grid-button": "grid-button button is-small",
        "grid-details-button": "grid-details-button button is-small",
        "grid-edit-button": "grid-edit-button button is-small",
        "grid-delete-button": "grid-delete-button button is-small",
        "grid-search-button": "grid-search-button button",
        "grid-clear-button": "grid-clear-button button",
        "grid-footer": "grid-footer pb-8",
        "grid-info": "grid-info is-pulled-left",
        "grid-pagination": "grid-pagination is-pulled-right",
        "grid-pagination-button": "grid-pagination-button button is-small",
        "grid-pagination-button-current": "grid-pagination-button-current button is-primary is-small",
        "grid-cell-type-string": "grid-cell-type-string",
        "grid-cell-type-text": "grid-cell-type-text",
        "grid-cell-type-boolean": "grid-cell-type-boolean has-text-centered",
        "grid-cell-type-float": "grid-cell-type-float",
        "grid-cell-type-decimal": "grid-cell-type-decimal",
        "grid-cell-type-int": "grid-cell-type-int",
        "grid-cell-type-date": "grid-cell-type-date",
        "grid-cell-type-time": "grid-cell-type-time",
        "grid-cell-type-datetime": "grid-cell-type-datetime",
        "grid-cell-type-upload": "grid-cell-type-upload",
        "grid-cell-type-list": "grid-cell-type-list",
        "grid-cell-type-id": "grid-cell-type-id has-text-centered",
        # specific for custom form
        "grid-search-form": "grid-search-form is-pulled-right pb-2",
        "grid-search-form-table": "grid-search-form-table",
        "grid-search-form-tr": "grid-search-form-tr",
        "grid-search-form-td": "grid-search-form-td pr-1",
        "grid-search-form-input": "grid-search-form-input input",
        "grid-search-form-select": "grid-search-form-input control select",
        "grid-search-form-error": "py4web-validation-error",
        "grid-search-boolean": "grid-search-boolean",
        "grid-header-element": "grid-header-element button",
        "grid-footer-element": "grid-footer-element button",
    }


class GridClassStyleBootstrap5(GridClassStyle):
    """The Grid style for Bootstrap 5"""

    classes = {
        "grid-wrapper": "grid-wrapper",
        "grid-header": "grid-header clearfix pb-2",
        "grid-new-button": "grid-new-button btn btn-outline-primary",
        "grid-search": "grid-search float-end pb-2",
        "grid-table-wrapper": "grid-table-wrapper table-responsive",
        "grid-table": "grid-table table table-striped table-hover table-bordered clearfix",
        "grid-sorter-icon-up": "grid-sort-icon-up fas fa-sort-up float-end",
        "grid-sorter-icon-down": "grid-sort-icon-down fas fa-sort-down float-end",
        "grid-thead": "",
        "grid-tr": "",
        "grid-th": "small",
        "grid-td": "small",
        "grid-td-buttons": "",
        "grid-back-button": "grid-button btn",
        "grid-button": "grid-button btn btn-sm btn-outline-secondary",
        "grid-details-button": "grid-details-button btn btn-sm btn-outline-secondary",
        "grid-edit-button": "grid-edit-button btn btn-sm btn-outline-secondary",
        "grid-delete-button": "grid-delete-button btn btn-sm btn-outline-secondary",
        "grid-search-button": "grid-search-button btn btn-outline-info",
        "grid-clear-button": "grid-clear-button btn btn-outline-info",
        "grid-footer": "grid-footer pb-8",
        "grid-info": "grid-info float-start",
        "grid-pagination": "grid-pagination float-end",
        "grid-pagination-button": "grid-pagination-button btn btn-sm btn-outline-primary",
        "grid-pagination-button-current": "grid-pagination-button-current btn btn-primary btn-sm",
        "grid-cell-type-string": "grid-cell-type-string",
        "grid-cell-type-text": "grid-cell-type-text",
        "grid-cell-type-boolean": "grid-cell-type-boolean text-center",
        "grid-cell-type-float": "grid-cell-type-float",
        "grid-cell-type-decimal": "grid-cell-type-decimal",
        "grid-cell-type-int": "grid-cell-type-int",
        "grid-cell-type-date": "grid-cell-type-date",
        "grid-cell-type-time": "grid-cell-type-time",
        "grid-cell-type-datetime": "grid-cell-type-datetime",
        "grid-cell-type-upload": "grid-cell-type-upload",
        "grid-cell-type-list": "grid-cell-type-list",
        "grid-cell-type-id": "grid-cell-type-id text-center",
        # specific for custom form
        "grid-search-form": "grid-search-form float-end pb-2",
        "grid-search-form-table": "grid-search-form-table",
        "grid-search-form-tr": "grid-search-form-tr",
        "grid-search-form-td": "grid-search-form-td pr-1",
        "grid-search-form-input": "grid-search-form-input form-control",
        "grid-search-form-select": "grid-search-form-input control select",
        "grid-search-form-error": "py4web-validation-error",
        "grid-search-boolean": "grid-search-boolean",
        "grid-header-element": "grid-header-element btn btn-sm",
        "grid-footer-element": "grid-footer-element btn btn-sm",
    }


class IconStyle:
    sort_up = "icon-sort-up"
    sort_down = "icon-sort-down"
    add_button = "icon-add-button"
    details_button = "icon-details-button"
    edit_button = "icon-edit-button"
    delete_button = "icon-delete-button"

    @classmethod
    def complete(self, icontxt: str) -> str:
        return icontxt


class IconStyleFontawsome(IconStyle):
    sort_up = "fas fa-sort-up"
    sort_down = "fas fa-sort-down"
    add_button = "fas fa-plus"
    details_button = "fas fa-id-card"
    edit_button = "fas fa-edit"
    delete_button = "fas fa-trash"

    @classmethod
    def complete(self, icontxt: str) -> str:
        if "fas " not in icontxt:
            return f"fas {icontxt}"
        return icontxt


class IconStyleBootstrapIcons(IconStyle):
    sort_up = "bi bi-sort-up"
    sort_down = "bi bi-sort-down"
    add_button = "bi bi-plus"
    details_button = "bi bi-person-vcard"
    edit_button = "bi bi-pencil-square"
    delete_button = "bi bi-trash"

    @classmethod
    def complete(self, icontxt: str) -> str:
        if "bi " not in icontxt:
            return f"bi {icontxt}"
        return icontxt


class GridClassStyleTailwind(GridClassStyle):
    classes = {
        "grid-wrapper": "grid-wrapper space-y-4 bg-white shadow-md rounded-lg p-4",
        "grid-header": "grid-header flex justify-between items-center pb-4 border-b border-gray-300",
        "grid-new-button": "grid-new-button px-4 py-2 bg-green-500 text-white rounded hover:bg-green-600",
        "grid-search": "grid-search flex items-center gap-2",
        "grid-table-wrapper": "grid-table-wrapper overflow-x-auto",
        "grid-table": "grid-table w-full border-collapse bg-white shadow-md rounded-lg overflow-hidden",
        "grid-sorter-icon-up": "grid-sort-icon-up fas fa-sort-up text-gray-600",
        "grid-sorter-icon-down": "grid-sort-icon-down fas fa-sort-down text-gray-600",
        "grid-thead": "bg-gray-200 text-gray-700",
        "grid-tr": "border-b border-gray-300",
        "grid-th": "px-4 py-2 text-left font-semibold",
        "grid-td": "px-4 py-2 text-gray-700",
        "grid-td-buttons": "px-4 py-2 flex gap-2",
        "grid-back-button": "px-3 py-1 bg-blue-500 text-white rounded hover:bg-blue-600 shadow",
        "grid-button": "px-3 py-1 bg-blue-500 text-white rounded hover:bg-blue-600 shadow",
        "grid-details-button": "px-3 py-1 bg-gray-500 text-white rounded hover:bg-gray-600 shadow",
        "grid-edit-button": "px-3 py-1 bg-yellow-500 text-white rounded hover:bg-yellow-600 shadow",
        "grid-delete-button": "px-3 py-1 bg-red-500 text-white rounded hover:bg-red-600 shadow",
        "grid-search-button": "px-4 py-2 bg-blue-500 text-white rounded hover:bg-blue-600 transition",
        "grid-clear-button": "px-4 py-2 bg-gray-500 text-white rounded hover:bg-gray-600 transition",
        "grid-footer": "grid-footer flex justify-between items-center pt-4 border-t border-gray-300",
        "grid-info": "grid-info text-gray-600",
        "grid-pagination": "grid-pagination flex gap-2",
        "grid-pagination-button": "px-3 py-1 bg-gray-200 rounded hover:bg-gray-300",
        "grid-pagination-button-current": "px-3 py-1 bg-blue-500 text-white rounded",
        # Cell styling
        "grid-cell-type-string": "text-left",
        "grid-cell-type-text": "text-left",
        "grid-cell-type-boolean": "text-center",
        "grid-cell-type-float": "text-right",
        "grid-cell-type-decimal": "text-right",
        "grid-cell-type-int": "text-right",
        "grid-cell-type-date": "text-center",
        "grid-cell-type-time": "text-center",
        "grid-cell-type-datetime": "text-center",
        "grid-cell-type-upload": "text-center",
        "grid-cell-type-list": "text-left",
        "grid-cell-type-id": "text-center",
        # Search form
        "grid-search-form": "flex flex-wrap gap-2 items-center border p-2 rounded-lg bg-gray-100",
        "grid-search-form-table": "w-full",
        "grid-search-form-tr": "border-b border-gray-300",
        "grid-search-form-td": "p-2",
        "grid-search-form-input": "px-3 py-2 border border-gray-300 rounded-md shadow-sm focus:ring-2 focus:ring-blue-500 focus:border-blue-500 w-full",
        "grid-search-form-select": "px-3 py-2 border border-gray-300 rounded-md shadow-sm focus:ring-2 focus:ring-blue-500 focus:border-blue-500 w-full",
        "grid-search-form-error": "p-2",
        "grid-search-boolean": "form-checkbox h-5 w-5 text-blue-600",
        "grid-header-element": "px-3 py-1 bg-gray-500 text-white rounded hover:bg-gray-600",
        "grid-footer-element": "px-3 py-1 bg-gray-500 text-white rounded hover:bg-gray-600",
    }


class Column:
    """class used to represent a column in a grid"""

    def __init__(
        self,
        name,
        represent,
        key=None,
        required_fields=None,  # must be a list or none
        orderby=None,
        col_type="string",
        td_class_style=None,
    ):
        self.name = name
        self.represent = represent
        self.orderby = orderby
        self.required_fields = required_fields or []
        self.key = key
        # col type is not quite a field type but we support field types
        self.type = col_type
        self.td_class_style = td_class_style


def maybe_call(obj, *args, **kwargs):
    """If the object is callable, call it with the args, kwargs"""
    if not callable(obj):
        return obj
    return obj(*args, **kwargs)


def reference_represent(field, value):
    """
    Assumes value is a pydal.objects.Reference value and represts is using the
    table._format of the referenced table
    """
    if not value:
        return ""
    table = field.referenced_table()
    if not table:
        return "#{value}"
    row = table(value)
    if not row:
        return "#{value}(missing)"
    if isinstance(table._format, str):
        return table._format % row
    elif callable(table._format):
        return table._format(row)
    return str(value)


def datetime_represent(value):
    """Represents a datetime value"""
    if not value or not isinstance(value, datetime.datetime):
        return value or ""
    return XML(
        "<script>document.currentScript.insertAdjacentText('afterend',"
        f"(new Date('{value.isoformat(timespec='seconds')}')).toLocaleString())</script>"
    )


def date_represent(value):
    """Represents a date value"""
    if not value or not isinstance(value, datetime.date):
        return value or ""
    return XML(
        "<script>document.currentScript.insertAdjacentText('afterend',"
        f"(new Date('{value.isoformat()}')).toLocaleDateString())</script>"
    )


def time_represent(value):
    """Represents a time value"""
    if not value or not isinstance(value, datetime.time):
        return value or ""
    return XML(
        "<script>document.currentScript.insertAdjacentText('afterend',"
        f"(new Date('0000T{value.isoformat('seconds')}')).toLocaleTimeString())</script>"
    )


class Grid:
    represent_by_type = {
        "id": lambda field, value: f"#{value}",
        "boolean": lambda field, value: "☑" if value else "☐" if value is False else "",
        "integer": lambda field, value: str(value) if value is not None else "",
        "bigint": lambda field, value: str(value) if value is not None else "",
        "float": lambda field, value: "%.2f" % value if value is not None else "",
        "double": lambda field, value: "%.2f" % value if value is not None else "",
        "decimal": lambda field, value: "%.2f" % value if value is not None else "",
        "reference": reference_represent,
        "big-reference": reference_represent,
        "datetime": lambda field, value: datetime_represent(value),
        "date": lambda field, value: date_represent(value),
        "time": lambda field, value: time_represent(value),
        "list": lambda field, value: ", ".join(str(x) for x in value) or "",
        "password": lambda field, value: "******",
    }

    def __init__(
        self,
        query,
        search_form=None,
        search_queries=None,
        columns=None,
        field_id=None,
        show_id=None,
        orderby=None,
        left=None,
        headings=None,
        create=True,
        details=True,
        editable=True,
        deletable=True,
        validation=None,
        required_fields=None,
        pre_action_buttons=None,
        post_action_buttons=None,
        auto_process=True,
        rows_per_page=15,
        include_action_button_text=True,
        search_button_text="Filter",
        formstyle=FormStyleDefault,
        grid_class_style=GridClassStyle,
        icon_style=IconStyleFontawsome,
        T=lambda text: text,
        groupby=None,
        # deprecated
        fields=None,
        form_maker=Form,
    ):
        """
        Grid is a searchable/sortable/pageable grid

        :param path: The part of the URL to be parsed by this grid
        :param query: the query used to filter the data
        :param search_form: py4web FORM to be included as the search form
        :param search_queries: future use - pass a dict of name and a search query
        :param columns: list of columns to display on the list page, if blank, glean tablename from first query
        :              and use all fields of that table
        :param field_id: the id field of the primary table for the grid - if there are multiple tables (joined)
                            then the table for this field is used as the table for edit/details/delete
        :param show_id: show the record id field on list page - default = False
        :param orderby: pydal orderby field or list of fields
        :param left: if joining other tables, specify the pydal left expression here
        :param headings: list of headings to be used for list page - if not provided use the field label
        :param create: URL to redirect to for creating records - set to False to not display the button
        :param editable: URL to redirect to for editing records - set to False to not display the button
        :param deletable: URL to redirect to for deleting records - set to False to not display the button
        :param validation: optional validation function to pass to create and edit forms
        :param pre_action_buttons: list of Column instances to include before the standard action buttons
        :param post_action_buttons: list of Column instances to include after the standard action buttons
        :param auto_process: True/False - automatically process the sql for the form - if False, user is
                              responsible for calling process().
        :param T: optional pluralize object
        """

        if show_id is None:
            show_id = fields and any(field.type == "id" for field in fields)

        self.param = Param(
            columns=columns or fields,
            field_id=field_id,
            show_id=show_id,
            orderby=orderby,
            left=left,
            groupby=groupby,
            search_form=search_form,
            search_queries=search_queries,
            headings=headings or [],
            create=create,
            details=details,
            editable=editable,
            deletable=deletable,
            validation=validation,
            pre_action_buttons=pre_action_buttons,
            post_action_buttons=post_action_buttons,
            rows_per_page=rows_per_page,
            include_action_button_text=include_action_button_text,
            search_button_text=search_button_text,
            formstyle=formstyle,
            grid_class_style=grid_class_style,
            new_sidecar=None,
            new_submit_value=None,
            new_action_button_text=T("New"),
            details_sidecar=None,
            details_submit_value=T("Back"),
            details_action_button_text=T("Details"),
            edit_sidecar=None,
            edit_submit_value=None,
            edit_action_button_text=T("Edit"),
            delete_action_button_text=T("Delete"),
            delete_submit_value=T("Confirm Delete"),
            back_button_value=T("Back"),
            header_elements=None,
            footer_elements=None,
            required_fields=required_fields or [],
            icon_style=icon_style,
        )

        # in case the query is a Table insteance
        if isinstance(query, query._db.Table):
            query = query._id != None

        #  instance variables that will be computed
        self.db = query._db  # the database
        self.query = query  # the filter query
        self.query_parms = safely(lambda: request.params, default={})
        self.search_query_error = None  # error to be displayed in case failed search
        self.T = T  # the translator
        self.form_maker = form_maker  # the object that makes forms
        self.this_url = None  # page referring this one
        self.mode = None  # select, new, details, edit, delete
        self.table = None  # selected table
        self.tablename = None  # name of the selected table
        self.total_number_of_rows = None  # total number of rows
        self.rows = None  # selected rows
        self.number_of_pages = None  # total number of pages for pagination
        self.current_page_number = None  # number of the current page
        self.page_end = None  # index of last element in page
        self.page_start = None  # index of first element in page
        self.record_id = None  # the record id to display or None
        self.record = None  # the record to display or None
        self.form = None  # the edit, new, details form
        self.hidden_fields = None  # hidden fields to be embedded in form
        self.represent_by_type = copy.copy(Grid.represent_by_type)

        if auto_process:
            self.process()

    def is_creatable(self):
        return False if self.param.groupby else maybe_call(self.param.create)

    def is_editable(self, row):
        return (
            False
            if not row or self.param.groupby
            else maybe_call(self.param.editable, row)
        )

    def is_readable(self, row):
        return False if not row else maybe_call(self.param.details, row)

    def is_deletable(self, row):
        return (
            False
            if not row or self.param.groupby
            else maybe_call(self.param.deletable, row)
        )

    def get_style(self, element, default=None):
        return self.param.grid_class_style.get(element, default)

    @staticmethod
    def parse(query):
        mode = query.get("mode", "select")
        record_id = query.get("id", None)
        if record_id and mode == "select":
            mode = "details"
        referrer = query.get("referrer")
        if referrer:
            referrer = base64.b16decode(referrer.encode("utf8")).decode("utf8")
        return {"mode": mode, "record_id": record_id, "referrer": referrer}

    def process(self):
        query = None
        db = self.db
        self.search_query_error = None

        parsed = Grid.parse(request.query)
        self.mode = parsed["mode"]
        self.record_id = parsed["record_id"]
        self.referrer = parsed["referrer"]
        # check invalid parameters]
        if self.mode not in ("new", "edit", "select", "details", "delete"):
            raise HTTP(400)
        elif self.mode in ("edit", "details", "delete") and self.record_id is None:
            raise HTTP(400)

        if self.param.field_id:
            self.tablename = str(self.param.field_id._table)
        else:
            self.tablename = self._get_tablenames(self.query)[0]
            self.param.field_id = db[self.tablename]._id

        if not self.tablename:
            raise HTTP(400)

        # SECURITY: if the record does not exist or does not match query, than we are not allowed
        self.table = db[self.tablename]

        # if we have a record id retrieve the record
        if self.record_id:
            self.record = self.table(self.record_id)
            if not self.record:
                redirect(URL())
        else:
            self.record = None

        #  ensure the user has access for new/details/edit/delete if chosen
        if self.mode == "select":
            self._handle_mode_select()
        elif self.mode == "new":
            self._handle_mode_new()
        elif self.mode == "details":
            self._handle_mode_details()
        elif self.mode == "edit":
            self._handle_mode_edit()
        elif self.mode == "delete":
            self._handle_mode_delete()

    def _handle_mode_new(self):
        if not self.is_creatable():
            raise HTTP(
                403,
                f"You do not have access to create a record in the {self.tablename} table.",
            )
        self.form = self._make_form(readonly=False)
        if self.param.new_sidecar:
            self.form.param.sidecar.append(self.param.new_sidecar)
        if self.param.new_submit_value:
            self.form.param.submit_value = self.param.new_submit_value
        self._handle_redirect_to_referrer(True, False)

    def _handle_mode_details(self):
        if not self.is_readable(self.record):
            raise HTTP(
                403,
                f"You do not have access to read a record from the {self.tablename} table.",
            )
        self.form = self._make_form(readonly=True)
        if self.param.details_sidecar:
            self.form.param.sidecar.append(self.param.details_sidecar)
        if self.param.details_submit_value:
            self.form.param.submit_value = self.param.details_submit_value
        self._handle_redirect_to_referrer(False, True)

    def _handle_mode_edit(self):
        if not self.is_editable(self.record):
            raise HTTP(
                403,
                f"You do not have access to edit a record in the {self.tablename} table.",
            )
        self.form = self._make_form(readonly=False)
        if self.param.edit_sidecar:
            self.form.param.sidecar.append(self.param.edit_sidecar)
        if self.param.edit_submit_value:
            self.form.param.submit_value = self.param.edit_submit_value
        self._handle_redirect_to_referrer(True, False)

    def _handle_mode_delete(self):
        if not self.is_deletable(self.record):
            self._handle_mode_delete()
        self.form = self._make_form(readonly=True)
        if self.param.delete_submit_value:
            self.form.param.submit_value = self.param.delete_submit_value
        if request.method == "POST":
            self.record.delete_record()
        self._handle_redirect_to_referrer(True, True)

    def _handle_redirect_to_referrer(self, inject_back, redirect_on_post):
        if self.referrer and inject_back:
            classes = self.get_style("grid-back-button")
            self.form.param.sidecar.insert(
                0,
                A(
                    self.param.back_button_value,
                    _role="button",
                    _class=classes,
                    _href=self.referrer,
                ),
            )

        # redirect to the referrer
        if self.form.accepted or (redirect_on_post and request.method == "POST"):
            redirect(self.referrer or URL())

    def _handle_mode_select(self):
        db = self.db
        query = self.query

        # use the default search for the table (we have to do this here because need self.table)
        if self.param.search_queries is None:
            self.param.search_queries = [make_default_search_query(self.table)]

        # apply the search query
        if not self.param.search_form and self.param.search_queries:
            search_type = safe_int(request.query.get("search_type", 0), default=0)
            search_string = request.query.get("search_string")
            if search_type < len(self.param.search_queries) and search_string:
                parts = self.param.search_queries[search_type]
                query_lambda = parts[1]
                if len(parts) == 3 and parts[2]:
                    requires = parts[2]
                    search_string, self.search_query_error = requires(search_string)
                if not self.search_query_error:
                    try:
                        query = self.query & query_lambda(search_string)
                    except Exception as e:
                        self.search_query_error = str(e)

        # if no column specified use all fields
        if not self.param.columns:
            self.param.columns = [field for field in self.table if field.readable]
        # convert to column object
        self.columns = []

        def title(col):
            return str(col).replace('"', "")

        def col2key(col):
            return str(col).lower().replace(".", "-")

        for index, col in enumerate(self.param.columns):
            if isinstance(col, Column):
                if not col.key:
                    col.key = f"column-{index}"
                self.columns.append(col)

            elif isinstance(col, Field):
                # deal with download links in special manner if no representation
                type_name = strip_field_type(col.type)
                # upload type is always special!
                if type_name == "upload" and hasattr(col, "download_url"):
                    represent_col = (
                        lambda row, name=str(col), f=col.download_url: row[name]
                        and A("download", _href=f(row[name]))
                        or ""
                    )
                # field represent override default representation by type
                elif col.represent:
                    represent_col = lambda row, name=str(col), f=col.represent: f(
                        row[name], row
                    )
                # we do not know better, use formarepresent by type (type is the stripped Field type)
                elif type_name in self.represent_by_type:
                    represent_col = lambda row, col=col, f=self.represent_by_type[
                        col.type_name
                    ]: f(col, row[str(col)])
                else:
                    represent_col = lambda row, name=str(col): row[name]

                self.columns.append(
                    Column(
                        col.label,
                        represent_col,
                        orderby=col,
                        required_fields=[col],
                        key=col2key(col),
                        col_type=type_name,
                    )
                )
            elif isinstance(col, FieldVirtual):

                def compute(row, col=col):
                    return col.f(row) if "id" in row else col.f(row[col.tablename])

                self.columns.append(
                    Column(
                        col.label,
                        compute,
                        orderby=None,
                        required_fields=db[col.tablename],
                        key=col2key(col),
                    )
                )
            elif isinstance(col, Expression):

                def compute(row, name=str(col)):
                    return row._extra(name)

                self.columns.append(
                    Column(
                        title(col),
                        compute,
                        orderby=None,
                        required_fields=[col],
                        key=f"column-{index}",
                    )
                )
            else:
                raise RuntimeError(f"Column not support {col}")

        # join the set of all required fields
        sets = [set(self.param.required_fields or [])]
        sets += [set(col.required_fields) for col in self.columns]
        self.needed_fields = list(
            functools.reduce(lambda a, b: a | b, sets) | set([self.table._id])
        )

        self.this_url = base64.b16encode(request.url.encode("utf8")).decode("utf8")
        self.current_page_number = safe_int(request.query.get("page"), default=1)

        select_params = dict()
        #  try getting sort order from the request
        sort_order = request.query.get("orderby")

        select_params["orderby"] = self.param.orderby
        if sort_order:
            parts = sort_order.lstrip("~").split(".")
            if len(parts) == 2 and parts[0] in db.tables and parts[1] in db[parts[0]]:
                orderby = db[parts[0]][parts[1]]
                if sort_order.startswith("~"):
                    orderby = ~orderby
                select_params["orderby"] = orderby

        if self.param.left:
            select_params["left"] = self.param.left

        if self.param.groupby:
            select_params["groupby"] = self.param.groupby

        if self.param.groupby or self.param.left:
            #  need groupby fields in select to get proper count
            self.total_number_of_rows = len(
                db(query).select(db[self.tablename]._id, **select_params)
            )
        else:
            self.total_number_of_rows = db(query).count()

        #  if at a high page number and then filter causes less records to be displayed, reset to page 1
        if (
            self.current_page_number - 1
        ) * self.param.rows_per_page > self.total_number_of_rows:
            self.current_page_number = 1

        if self.total_number_of_rows > self.param.rows_per_page:
            self.page_start = self.param.rows_per_page * (self.current_page_number - 1)
            self.page_end = self.page_start + self.param.rows_per_page
            select_params["limitby"] = (self.page_start, self.page_end)
        else:
            self.page_start = 0
            if self.total_number_of_rows > 1:
                self.page_start = 1
            self.page_end = self.total_number_of_rows

        # get the data
        self.rows = db(query).select(*self.needed_fields, **select_params)

        self.number_of_pages = self.total_number_of_rows // self.param.rows_per_page
        if self.total_number_of_rows % self.param.rows_per_page > 0:
            self.number_of_pages += 1

        if (
            self.param.pre_action_buttons
            or self.param.details
            or self.param.editable
            or self.param.deletable
            or self.param.post_action_buttons
        ):
            key = f"column-{len(self.columns)}"
            self.columns.append(
                Column(
                    "",
                    self.make_action_buttons,
                    key=key,
                    td_class_style=self.get_style("grid-td-buttons"),
                )
            )

    def _make_form(self, readonly):
        return self.form_maker(
            self.table,
            record=self.record,
            readonly=readonly,
            deletable=self.is_deletable(self.record),
            formstyle=self.param.formstyle,
            validation=self.param.validation,
            show_id=self.param.show_id,
        )

    def _iter_pages(
        self,
        current_page,
        num_pages,
        left_current=1,
        right_current=1,
        left_edge=1,
        right_edge=1,
    ):
        """
        generator used to determine which page numbers should be shown on the Grid pager

        :param left_current: # of pages to add to the left of current page
        :param right_current: # of fpages to add to the right of current page
        :param left_edge: # of pages to add to the left end
        :param right_edge: # of fpages to add to the right end
        """
        left_range = set(range(1, min(2 + left_edge, num_pages + 1)))
        mid_range = set(
            range(
                max(1, current_page - left_current),
                min(current_page + right_current + 1, num_pages + 1),
            )
        )
        right_range = set(range(max(1, num_pages - right_edge), num_pages + 1))
        return list(sorted(left_range | mid_range | right_range))

    def _make_action_button(
        self,
        text,
        url,
        icon=None,
        classes=None,
        kind="grid-button",
        **attrs,
    ):
        if kind:
            classes = join_classes(classes, self.get_style(kind))

        if not "_role" in attrs:
            attrs["_role"] = "button"
        link = A(
            I(_class=self.param.icon_style.complete(icon)) if icon else "",
            _class=classes,
            _href=url,
            **attrs,
        )
        if self.param.include_action_button_text:
            link.append(
                SPAN(
                    XML("&nbsp;"),
                    text,
                    _class=self.get_style(kind + "-text") if kind else None,
                )
            )

        return link

    def _make_default_form(self):
        search_type = safe_int(request.query.get("search_type", 0), default=0)
        search_string = request.query.get("search_string")
        options = [
            OPTION(items[0], _value=k, _selected=(k == search_type))
            for k, items in enumerate(self.param.search_queries)
        ]
        hidden_fields = [
            INPUT(_name=key, _value=request.query.get(key), _type="hidden")
            for key in request.query
            if key not in ("search_type", "search_string")
        ]
        form = FORM(*hidden_fields, _class=self.get_style("grid-search-form"))
        select = SELECT(
            *options,
            _name="search_type",
            _class=self.get_style("grid-search-form-select"),
        )
        input = INPUT(
            _type="text",
            _name="search_string",
            _value=search_string,
            _class=self.get_style("grid-search-form-input"),
        )
        submit = INPUT(
            _type="submit",
            _value=self.T("Search"),
            _class=self.get_style("grid-search-button"),
        )
        clear_script = "document.querySelector('[name=search_string]').value='';"
        clear = INPUT(
            _type="submit",
            _value=self.T("Clear"),
            _onclick=clear_script,
            _class=self.get_style("grid-clear-button"),
        )
        div = DIV(_id="grid-search", _classes=self.get_style("grid-search"))

        tr = TR(_class=self.get_style("grid-search-form-tr"))
        td_classes = self.get_style("grid-search-form-td")
        if len(options) > 1:
            tr.append(TD(select, _class=td_classes))
        tr.append(TD(input, _class=td_classes))
        tr.append(TD(submit, clear, _class=td_classes))
        table = TABLE(tr, _class=self.get_style("grid-search-form-table"))
        if self.search_query_error:
            table.append(
                TR(
                    TD(
                        self.search_query_error,
                        _colspan=3 if len(options) > 1 else 2,
                        _class=self.get_style("grid-search-form-error"),
                    )
                )
            )
        form.append(table)
        div.append(form)
        return div

    def _make_search_form(self):
        # TODO: Do we need this?
        div = DIV(_id="grid-search", _class=self.get_style("grid-search"))
        div.append(self.param.search_form.custom["begin"])
        tr = TR(_class=self.get_style("grid-search-form-tr"))
        for field in self.param.search_form.table:
            td = TD(_class=self.get_style("grid-search-form-td"))
            if field.type == "boolean":
                sb = DIV(_class=self.get_style("grid-search-boolean"))
                sb.append(self.param.search_form.custom["widgets"][field.name])
                sb.append(field.label)
                td.append(sb)
            else:
                td.append(self.param.search_form.custom["wrappers"][field.name])
            if (
                field.name in self.param.search_form.custom["errors"]
                and self.param.search_form.custom["errors"][field.name]
            ):
                td.append(DIV(self.param.search_form.custom["errors"][field.name]))
            tr.append(td)
        if self.param.search_button_text:
            tr.append(
                TD(
                    INPUT(
                        _class="button",
                        _type="submit",
                        _value=self.T(self.param.search_button_text),
                    ),
                    _class=self.get_style("grid-search-form-td"),
                )
            )
        else:
            tr.append(
                TD(
                    self.param.search_form.custom["submit"],
                    _class=self.get_style("grid-search-form-td"),
                )
            )
        div.append(TABLE(tr, _class=self.get_style("grid-search-form-table")))
        for hidden_widget in self.param.search_form.custom["hidden_widgets"].keys():
            if hidden_widget not in ("formname", "formkey"):
                div.append(
                    self.param.search_form.custom["hidden_widgets"][hidden_widget]
                )

        div.append(self.param.search_form.custom["end"])

        return div

    def _make_table_header(self):
        sort_order = request.query.get("orderby", "")

        thead = THEAD(_class=self.get_style("grid-thead"))
        for index, col in enumerate(self.columns):
            col_header = self._make_col_header(col, index, sort_order)
            classes = join_classes(
                self.get_style("grid-th"),
                f"grid-col-{col.key}",
            )
            thead.append(TH(col_header, _class=classes))

        return thead

    def _make_col_header(self, col, index, sort_order):
        up = I(
            _class=join_classes(
                self.get_style("grid-sorter-icon-up"), self.param.icon_style.sort_up
            )
        )
        dw = I(
            _class=join_classes(
                self.get_style("grid-sorter-icon-down"), self.param.icon_style.sort_down
            )
        )

        orderby = col.orderby and str(col.orderby)

        heading = (
            self.param.headings[index] if index < len(self.param.headings) else col.name
        )
        #  add the sort order query parm
        sort_query_parms = dict(self.query_parms)

        attrs = {}
        if orderby:
            if orderby == sort_order:
                sort_query_parms["orderby"] = "~" + orderby
                url = URL(vars=sort_query_parms)
                col_header = A(heading, up, _href=url)
            else:
                sort_query_parms["orderby"] = orderby
                url = URL(vars=sort_query_parms)
                col_header = A(
                    heading, dw if "~" + orderby == sort_order else "", _href=url
                )
        else:
            col_header = heading
        return col_header

    def _make_table_body(self):
        tbody = TBODY()
        for row in self.rows:
            #  find the row id - there may be nested tables....

            tr = TR(_role="row", _class=self.get_style("grid-tr"))

            #  add all the fields to the row
            for col in self.columns:
                classes = join_classes(
                    [
                        self.get_style(maybe_call(col.td_class_style, row), "grid-td"),
                        f"grid-cell-{col.key}",
                    ]
                )
                value = col.represent(row)
                tr.append(TD(value, _class=classes))

            tbody.append(tr)

        return tbody

    def make_action_buttons(self, row):
        cat = CAT()
        row_id = row[self.param.field_id] if self.param.field_id else row.id
        if self.param.pre_action_buttons and len(self.param.pre_action_buttons) > 0:
            for btn in self.param.pre_action_buttons:
                btn = maybe_call(btn, row)
                if btn is None:
                    # if None, no button
                    continue
                if isinstance(btn, dict):
                    cat.append(self._make_action_button(**btn))
                else:
                    cat.append(btn)

        if self.is_readable(row):
            if isinstance(self.param.details, str):
                details_url = self.param.details.format(id=row_id)
            else:
                details_url = URL(vars=dict(id=row_id, referrer=self.this_url))
            cat.append(
                self._make_action_button(
                    text=self.T(self.param.details_action_button_text),
                    url=details_url,
                    icon=self.param.icon_style.details_button,
                    kind="grid-details-button",
                )
            )
        if self.is_editable(row):
            if isinstance(self.param.editable, str):
                edit_url = self.param.editable.format(id=row_id)
            else:
                edit_url = URL(
                    vars=dict(mode="edit", id=row_id, referrer=self.this_url)
                )
            cat.append(
                self._make_action_button(
                    text=self.T(self.param.edit_action_button_text),
                    url=edit_url,
                    icon=self.param.icon_style.edit_button,
                    kind="grid-edit-button",
                    _disabled=not self.is_editable(row),
                )
            )
        if self.is_deletable(row):
            if isinstance(self.param.deletable, str):
                delete_url = self.param.deletable.format(id=row_id)
            else:
                delete_url = URL(
                    vars=dict(mode="delete", id=row_id, referrer=self.this_url)
                )
            cat.append(
                self._make_action_button(
                    text=self.T(self.param.delete_action_button_text),
                    url=delete_url,
                    icon=self.param.icon_style.delete_button,
                    additional_classes="confirmation",
                    kind="grid-delete-button",
                    _disabled=not self.is_deletable(row),
                )
            )

        if self.param.post_action_buttons and len(self.param.post_action_buttons) > 0:
            for btn in self.param.post_action_buttons:
                btn = maybe_call(btn, row)
                if btn is None:
                    # if None, no button
                    continue
                if isinstance(btn, dict):
                    cat.append(self._make_action_button(**btn))
                else:
                    cat.append(btn)

        return cat

    def _make_table_pager(self):
        pager = DIV(_class=self.get_style("grid-pagination"))
        previous_page_number = None
        for page_number in self._iter_pages(
            self.current_page_number, self.number_of_pages
        ):
            pager_query_parms = dict(self.query_parms)
            pager_query_parms["page"] = page_number
            # if there is a gat add a spacer
            if previous_page_number and page_number - previous_page_number > 1:
                pager.append(SPAN("..."))
            is_current = self.current_page_number == page_number
            page_name = (
                "grid-pagination-button-current"
                if is_current
                else "grid-pagination-button"
            )
            pager.append(
                A(
                    page_number,
                    _class=self.get_style(page_name),
                    _role="button",
                    _href=URL(vars=pager_query_parms),
                )
            )
            previous_page_number = page_number
        return pager

    def _make_table(self):
        html = DIV(_class=self.get_style("grid-wrapper"))
        grid_header = DIV(_class=self.get_style("grid-header"))

        #  build the New button if needed
        if self.param.create and self.param.create != "":
            if isinstance(self.param.create, str):
                create_url = self.param.create
            else:
                create_url = URL(vars=dict(mode="new", referrer=self.this_url))

            grid_header.append(
                self._make_action_button(
                    self.T(self.param.new_action_button_text),
                    create_url,
                    icon=self.param.icon_style.add_button,
                    kind="grid-new-button",
                )
            )
        if self.param.header_elements and len(self.param.header_elements) > 0:
            for element in self.param.header_elements:
                element = maybe_call(element)
                if isinstance(element, dict):
                    element = copy.copy(element)
                    element["kind"] = element.get("kind", "grid-header-element")
                    grid_header.append(self._make_action_button(**element))
                else:
                    grid_header.append(element)

        #  build the search form if provided
        if self.param.search_form:
            grid_header.append(self._make_search_form())
        elif self.param.search_queries and len(self.param.search_queries) > 0:
            grid_header.append(self._make_default_form())

        html.append(grid_header)

        table = TABLE(_class=self.get_style("grid-table"))

        # build the header
        table.append(self._make_table_header())

        #  build the rows
        table.append(self._make_table_body())

        #  add the table to the html
        html.append(DIV(table, _class=self.get_style("grid-table-wrapper")))

        #  add the row counter information
        footer = DIV(_class=self.get_style("grid-footer"))

        row_count = DIV(_class=self.get_style("grid-info"))
        (
            row_count.append(
                str(self.T("Displaying rows %s thru %s of %s"))
                % (
                    self.page_start + 1 if self.number_of_pages > 1 else 1,
                    (
                        self.page_end
                        if self.page_end < self.total_number_of_rows
                        else self.total_number_of_rows
                    ),
                    self.total_number_of_rows,
                )
            )
            if self.number_of_pages > 0
            else row_count.append(self.T("No rows to display"))
        )
        footer.append(row_count)

        #  build the pager
        if self.number_of_pages > 1:
            footer.append(self._make_table_pager())

        if self.param.footer_elements and len(self.param.footer_elements) > 0:
            for element in self.param.footer_elements:
                element = maybe_call(element)
                if isinstance(element, dict):
                    element = copy.copy(element)
                    element["kind"] = element.get("kind", "grid-footer-element")
                    footer.append(self._make_action_button(**element))
                else:
                    footer.append(element)

        html.append(footer)
        return html

    def render(self):
        """
        build the query table

        :return: html representation of the table or the py4web Form object
        """
        if self.mode == "select":
            return self._make_table()
        elif self.mode in ["new", "details", "edit", "delete"]:
            return self.form
        raise HTTP(404)

    def xml(self):
        return self.render().xml()

    def __str__(self):
        return str(self.xml())

    def data(self):
        """
        get the record that is being edited / displayed

        :return: DAL record of the record being edited
        """
        return (
            self.db[self.tablename](self.record_id)
            if self.tablename and self.record_id
            else None
        )

    def _get_tablenames(self, *args):
        """Returns the tablenames used by this grid"""
        return list(self.db._adapter.tables(*args).keys())

    def _is_join(self):
        """is this a left join?"""
        items = [self.query]
        if self.param.left is not None:
            if isinstance(self.param.left, (list, tuple)):
                items += [item for item in self.param.left]
            else:
                items += [self.param.left]
        return len(self._get_tablenames(*items)) > 1


def parse_referrer(r):
    """
    Get the referrer from the request query and use urlparse to parse it into a url object

    :param r: The request object to be interrogated

    :return: the URL object or None if no URL object obtained/decoded
    """
    referrer = r.query.get("referrer")
    url = None
    if referrer:
        url_string = base64.b16decode(referrer.encode("utf8")).decode("utf8")
        if url_string:
            url = urlparse(url_string)

    return url


def get_parent(parent_field):
    """
    try to find the parent id for a parent/child table relationship

    :param path: the path var from the method calling the grid
    :param child_table: the child table of the parent/child relationship ex. db.child_table
    :param parent_field_name: this is the name of the field in the child table that points to the parent table
                              this is a string value

    :return parent_id: the id of the parent record
    """
    parent_id = request.query.get("parent_id")
    if parent_id is not None:
        return int(parent_id)

    child_table = parent_field._table
    fn = parent_field.name

    #  if not found, search the record id of the parent from the child table record
    record_id = request.query.get("id")
    if record_id:
        record = child_table(record_id)
        if record:
            parent_id = record[fn]
            if parent_id is not None:
                return int(parent_id)

    #  else, check in the form
    parent_id = request.forms.get(fn)
    if parent_id is not None:
        return int(parent_id)

    #  else, check in the referrer
    referrer = request.query.get("referrer")
    if referrer:
        kvp = base64.b16decode(referrer.encode("utf8")).decode("utf8")
        if "parent_id" in kvp:
            parent_id = kvp.split("parent_id=")[1].split("&")[0]
            return int(parent_id)

    return None



import io
import os
import unittest
import uuid

from py4web.core import DAL, HTTP, Field, Fixture, Session, bottle, request, safely
from py4web.utils.auth import Auth, AuthAPI

SECRET = str(uuid.uuid4())


class TestAuth(unittest.TestCase):
    def setUp(self):
        os.environ["PY4WEB_APPS_FOLDER"] = "apps"
        self.db = DAL("sqlite:memory")
        self.session = Session(secret=SECRET, expiration=10)
        self.auth = Auth(
            self.session, self.db, define_tables=True, password_complexity=None
        )
        self.auth.enable()
        self.auth.action = self.action
        request.app_name = "_scaffold"

    def tearDown(self):
        # this is normally done by @action
        safely(lambda: Fixture.local_delete(self.session))
        bottle.app.router.remove("/*")

    def action(self, name, method, query, data):
        request.environ["REQUEST_METHOD"] = method
        request.environ["ombott.request.query"] = query
        request.environ["ombott.request.json"] = data
        request.environ["wsgi.input"] = io.BytesIO()
        # we break a symmetry below. should fix in auth.py
        if name.startswith("api/"):
            return getattr(AuthAPI, name[4:])(self.auth)
        return getattr(self.auth.form_source, name)()

    def on_request(self, context=None, keep_session=False):
        # store the current session
        context = context or {}
        try:
            storage = self.session.local.__dict__
        except RuntimeError:
            storage = None
        # reinitialize everything
        safely(lambda: Fixture.local_delete(self.session))
        safely(lambda: Fixture.local_delete(self.auth.flash))
        self.session.on_request(context)
        self.auth.flash.on_request(context)
        self.auth.on_request(context)
        # restore the previous session
        if keep_session and storage:
            self.session.local.__dict__.update(storage)

    def test_extra_fields(self):
        db = DAL("sqlite:memory")
        self.auth = Auth(
            self.session, db, define_tables=True, extra_fields=[Field("favorite_color")]
        )
        self.on_request()
        self.assertEqual(type(db.auth_user.favorite_color), Field)

    def test_register_invalid(self):
        self.on_request()
        body = {"email": "pinco.pallino@example.com"}
        res = self.auth.action("api/register", "POST", {}, body)
        self.assertEqual(
            res,
            {
                "id": None,
                "errors": {
                    "username": "required",
                    "password": "required",
                    "first_name": "required",
                    "last_name": "required",
                },
                "status": "error",
                "message": "validation errors",
                "code": 401,
            },
        )

    def test_register(self):
        self.on_request()
        body = {
            "username": "ppallino",
            "email": "pinco.pallino@example.com",
            "password": "123456789",
            "first_name": "Pinco",
            "last_name": "Pallino",
        }
        self.assertEqual(
            self.auth.action("api/register", "POST", {}, body),
            {"id": 1, "status": "success", "code": 200},
        )
        user = self.db.auth_user[1]
        self.assertTrue(user.action_token.startswith("pending-registration"))
        self.assertEqual(self.auth.get_user(), {})

        self.on_request()
        body = {"email": "pinco.pallino@example.com", "password": "1234567"}
        self.assertEqual(
            self.auth.action("api/login", "POST", {}, body),
            {"status": "error", "message": "Invalid Credentials", "code": 400},
        )

        self.on_request()
        body = {"email": "pinco.pallino@example.com", "password": "123456789"}
        self.assertEqual(
            self.auth.action("api/login", "POST", {}, body),
            {"status": "error", "message": "Registration is pending", "code": 400},
        )

        self.on_request()
        token = user.action_token[len("pending-registration") + 1 :]
        try:
            self.auth.action("verify_email", "GET", {"token": token}, {})
            assert False, "email not verified"
        except HTTP:
            pass
        user = self.db.auth_user[1]
        self.assertTrue(user.action_token is None)

        self.on_request()
        body = {"email": "pinco.pallino@example.com", "password": "1234567"}
        self.assertEqual(
            self.auth.action("api/login", "POST", {}, body),
            {"status": "error", "message": "Invalid Credentials", "code": 400},
        )

        self.on_request()
        body = {"email": "pinco.pallino@example.com", "password": "123456789"}
        self.assertEqual(
            self.auth.action("api/login", "POST", {}, body),
            {
                "user": {
                    "id": 1,
                    "username": "ppallino",
                    "email": "pinco.pallino@example.com",
                    "first_name": "Pinco",
                    "last_name": "Pallino",
                },
                "status": "success",
                "code": 200,
            },
        )

        self.on_request()
        body = {
            "email": "ppallino",  # can login with both email and username
            "password": "123456789",
        }
        self.assertEqual(
            self.auth.action("api/login", "POST", {}, body),
            {
                "user": {
                    "id": 1,
                    "username": "ppallino",
                    "email": "pinco.pallino@example.com",
                    "first_name": "Pinco",
                    "last_name": "Pallino",
                },
                "status": "success",
                "code": 200,
            },
        )

        self.on_request(keep_session=True)
        body = {"email": "pinco.pallino@example.com"}
        self.assertEqual(
            self.auth.action("api/request_reset_password", "POST", {}, body),
            {"status": "success", "code": 200},
        )

        self.on_request(keep_session=True)
        body = {"token": "junk", "new_password": "987654321"}
        self.assertEqual(
            self.auth.action("api/reset_password", "POST", {}, body),
            {
                "status": "error",
                "message": "validation errors",
                "errors": {"token": "invalid token"},
                "code": 401,
            },
        )

        self.on_request(keep_session=True)
        body = {
            "token": self.auth._link.split("?token=")[1],
            "new_password": "987654321",
            "new_password2": "987654321",
        }
        self.assertEqual(
            self.auth.action("api/reset_password", "POST", {}, body),
            {"status": "success", "code": 200},
        )

        self.assertEqual(
            self.auth.get_user(),
            {
                "id": 1,
                "username": "ppallino",
                "email": "pinco.pallino@example.com",
                "first_name": "Pinco",
                "last_name": "Pallino",
            },
        )

        self.on_request(keep_session=True)
        body = {}
        self.assertEqual(
            self.auth.action("api/change_password", "POST", {}, body),
            {
                "errors": {"old_password": "invalid current password"},
                "status": "error",
                "message": "validation errors",
                "code": 401,
            },
        )

        self.on_request(keep_session=True)
        body = {"old_password": "987654321", "new_password": "432187659"}
        self.assertEqual(
            self.auth.action("api/change_password", "POST", {}, body),
            {"updated": 1, "status": "success", "code": 200},
        )

        self.on_request(keep_session=True)
        body = {"password": "432187659", "new_email": "somebody@example.com"}
        self.assertEqual(
            self.auth.action("api/change_email", "POST", {}, body),
            {"updated": 1, "status": "success", "code": 200},
        )

        self.on_request(keep_session=True)
        body = {"first_name": "Max", "last_name": "Powers"}
        self.assertEqual(
            self.auth.action("api/profile", "POST", {}, body),
            {"updated": 1, "status": "success", "code": 200},
        )


if __name__ == "__main__":
    unittest.main()



import base64

from pydal._compat import urllib2, urlopen


class BasicAuthPlugin:
    name = "basic"
    label = "Basic"

    def __init__(self, server="127.0.0.1", table=None):
        self.server = server

    def check_credentials(self, username, password):
        """
        to use basic login with a different server
        from gluon.contrib.login_methods.basic_auth import basic_auth
        auth.settings.login_methods.append(basic_auth('http://server'))
        """
        key = base64.b64encode(username + ":" + password)
        headers = {"Authorization": "Basic " + key}
        request = urllib2.Request(self.server, None, headers)
        try:
            urlopen(request)
            return True
        except (urllib2.URLError, urllib2.HTTPError):
            return False



"builds and returns a wsgiref threading server"
# https://www.electricmonk.nl/log/2016/02/15/multithreaded-dev-web-server-for-the-python-bottle-web-framework/

import socket
import ssl
from concurrent.futures import ThreadPoolExecutor  # pip install futures
from socketserver import ThreadingMixIn
from wsgiref.simple_server import WSGIRequestHandler, WSGIServer, make_server

from ombott.server_adapters import ServerAdapter

from .logging_utils import logging_conf


class WSGIRefAdapter(ServerAdapter):
    "Class implementing a WSGIRef server"

    def run(self, handler):
        "runs the server"

        self.log = None

        if not self.quiet:
            self.log = logging_conf(
                self.options["logging_level"],
                "wsgiref",
            )

        self_run = self  # used in internal classes to access options and logger

        class LogHandler(WSGIRequestHandler):
            def address_string(self):  # Prevent reverse DNS lookups please.
                return self.client_address[0]

            def log_request(self, *args, **kw):
                if not self_run.quiet:
                    WSGIRequestHandler.log_request(self, *args, **kw)

            def log_message(self, formatstr, *args):
                if not self_run.quiet:  # and ( not args[1] in ['200', '304']) :
                    msg = "%s - - [%s] %s" % (
                        self.client_address[0],
                        self.log_date_time_string(),
                        formatstr % args,
                    )
                    self_run.log.info(msg)

        class PoolMixIn(ThreadingMixIn):
            def process_request(self, request, client_address):
                self.pool.submit(self.process_request_thread, request, client_address)

        class ThreadingWSGIServer(PoolMixIn, WSGIServer):
            daemon_threads = True
            pool = ThreadPoolExecutor(max_workers=self.options.get("workers", 40))

        class Server:
            def __init__(self, server_address=("127.0.0.1", 8000), handler_cls=None):
                self.wsgi_app = None
                self.listen, self.port = server_address
                self.handler_cls = handler_cls

            def set_app(self, app):
                self.wsgi_app = app

            def get_app(self):
                return self.wsgi_app

            def serve_forever(self):
                self.server = make_server(
                    self.listen,
                    self.port,
                    self.wsgi_app,
                    ThreadingWSGIServer,
                    self.handler_cls,
                )

                # openssl req -newkey rsa:4096 -new -x509 -keyout server.pem -out server.pem -days 365 -nodes
                # openssl req -x509 -newkey rsa:4096 -nodes -out cert.pem -keyout key.pem -days 365
                # ./py4web.py run apps -s wsgirefThreadingServer --watch=off --port=8443 --ssl_cert=cert.pem --ssl_key=key.pem
                # openssl s_client -showcerts -connect 127.0.0.1:8443

                certfile = self_run.options.get("certfile", None)

                if certfile:
                    self.server.socket = ssl.wrap_socket(
                        self.server.socket,
                        certfile=certfile,
                        keyfile=self_run.options.get("keyfile", None),
                        ssl_version=ssl.PROTOCOL_SSLv23,
                        server_side=True,
                        do_handshake_on_connect=False,
                    )

                self.server.serve_forever()

        server_cls = Server

        if ":" in self.host:  # Fix wsgiref for IPv6 addresses.
            if getattr(server_cls, "address_family") == socket.AF_INET:

                class ServerClass(Server):
                    address_family = socket.AF_INET6

                server_cls = ServerClass

        srv = make_server(
            self.host, self.port, handler, server_cls, LogHandler
        )  # handler_cls)
        srv.serve_forever()



# from geventwebsocket.handler import WebSocketHandler # pip install gevent-websocket
from gevent import pywsgi
from gevent_ws import WebSocketHandler  # pip install gevent gevent-ws
from ombott.server_adapters import ServerAdapter

from .logging_utils import logging_conf

# https://stackoverflow.com/questions/5312311/secure-websockets-with-self-signed-certificate
# https://pypi.org/project/gevent-ws/
# ./py4web.py run apps -s geventWebSocketServer --watch=off --ssl_cert=server.pem -H 192.168.1.161 -P 9000 -L 10
# vi apps/_websocket/templates/index.html    set: ws, wss, host, port
# firefox http://localhost:8000/_websocket
# firefox https://192.168.1.161:9000/_websocket  test wss
# curl --insecure -I -H 'Upgrade: websocket' \
#   -H "Sec-WebSocket-Key: `openssl rand -base64 16`" \
#   -H 'Sec-WebSocket-Version: 13' \
#   -sSv  https://192.168.1.161:9000/


class GeventWebSocketAdapter(ServerAdapter):
    "Class implementing a Gevent websocket server"

    def run(self, handler):
        "Runs the server"
        logger = None  # "default"

        if not self.quiet:
            logger = logging_conf(
                self.options["logging_level"],
                "gevent-ws",
            )

        args = dict(
            handler_class=WebSocketHandler,
            log=logger,
            error_log=logger,
        )

        certfile = self.options.get("certfile")
        keyfile = self.options.get("keyfile")
        if certfile and keyfile:
            args.update(certfile=certfile, keyfile=keyfile)

        server = pywsgi.WSGIServer((self.host, self.port), handler, **args)

        server.serve_forever()



# -*- coding: utf-8 -*-
#
# last tinkered with by korylprince at gmail.com on 2012-07-12
#

import logging
import sys

import ldap  # python-ldap
import ldap.filter

from . import UsernamePassword

ldap.set_option(ldap.OPT_REFERRALS, 0)


class LDAPPlugin(UsernamePassword):
    name = "ldap"
    label = "LDAP"

    """
    to use ldap login with MS Active Directory:

          auth.register_plugin(LDAPPlugin(
            mode='ad', server='my.domain.controller',
            base_dn='ou=Users,dc=domain,dc=com'))

    to use ldap login with Notes Domino:

          auth.register_plugin(LDAPPlugin(
            mode='domino',server='my.domino.server'))

    to use ldap login with OpenLDAP:

        auth.register_plugin(LDAPPlugin(
            mode='cn', server='my.ldap.server', base_dn='ou=Users,dc=domain,dc=com'))

    to use ldap login with OpenLDAP and subtree search and (optionally)
    multiple DNs:

        auth.register_plugin(LDAPPlugin(
            mode='uid_r', server='my.ldap.server',
            base_dn=['ou=Users,dc=domain,dc=com','ou=Staff,dc=domain,dc=com']))

    or (if using CN):

        auth.register_plugin(LDAPPlugin(
            mode='cn', server='my.ldap.server',
            base_dn='ou=Users,dc=domain,dc=com'))

    or you can full customize the search for user:

        auth.register_plugin(LDAPPlugin(
            mode='custom', server='my.ldap.server',
            base_dn='ou=Users,dc=domain,dc=com',
            username_attrib='uid',
            custom_scope='subtree'))

    the custom_scope can be: base, onelevel, subtree.

    If using secure ldaps:// pass secure=True and cert_path="..."
    If ldap is using GnuTLS then you need cert_file="..." instead cert_path
    because cert_path isn't implemented in GnuTLS :(

    To enable TLS, set tls=True:

        auth.register_plugin(LDAPPlugin(
            server='my.ldap.server',
            base_dn='ou=Users,dc=domain,dc=com',
            tls=True))

    If you need to bind to the directory with a bind account in order to
    search it, then specify bind_dn & bind_pw to use for this.
    - currently only implemented for Active Directory where anonymous bind
    is normally not allowed.

    If you need to restrict the set of allowed users (e.g. to members of a
    department) then specify an rfc4515 search filter string.
    - currently only implemented for mode in ['ad', 'company', 'uid_r']

    You can manage user attributes first name, last name, email from ldap:

        auth.register_plugin(LDAPPlugin(
            ...as usual...,
            manage_user=True,
            user_firstname_attrib='cn:1',
            user_lastname_attrib='cn:2',
            user_mail_attrib='mail'
           ))

    Where:
    manage_user: bool
        If True py4web will fetch and update user profile
        fields (first name, last name, email) from LDAP/AD on each login and
        keep them in sync with its db.
        If False, only authentication is performed and user profile fields
        are taken only from py4web db.
    user_firstname_attrib - the attribute containing the user's first name
                            optionally you can specify parts.
                            Example: cn: "John Smith" - 'cn:1'='John'
    user_lastname_attrib - the attribute containing the user's last name
                            optionally you can specify parts.
                            Example: cn: "John Smith" - 'cn:2'='Smith'
    user_mail_attrib - the attribute containing the user's email address


    If you need group control from ldap to py4web app's database feel free
    to set:

        auth.register_plugin(LDAPPlugin(
            ...as usual...,
            manage_groups=True,
            db=db,
            group_dn='ou=Groups,dc=domain,dc=com',
            group_name_attrib='cn',
            group_member_attrib='memberUid',
            group_filterstr='objectClass=*'
           ))

        Where:
        manage_groups - let py4web handle the groups from ldap
        db - is the database object (need to have auth_user, auth_group,
            auth_membership)
        group_dn - the ldap branch of the groups
        group_name_attrib - the attribute where the group name is stored
        group_member_attrib - the attribute containing the group members name
        group_filterstr - as the filterstr but for group select

    **allowed_groups still to be implemented in py4web**
    You can restrict login access to specific groups if you specify:

        auth.register_plugin(LDAPPlugin(
            ...as usual...,
            allowed_groups=[...],
            group_dn='ou=Groups,dc=domain,dc=com',
            group_name_attrib='cn',
            group_member_attrib='memberUid',#use 'member' for Active Directory
            group_filterstr='objectClass=*'
           ))

        Where:
        allowed_groups - a list with allowed ldap group names
        group_dn - the ldap branch of the groups
        group_name_attrib - the attribute where the group name is stored
        group_member_attrib - the attribute containing the group members name
        group_filterstr - as the filterstr but for group select

    If using Active Directory you must specify bind_dn and bind_pw for
    allowed_groups because anonymous bind is not normally allowed.

    """

    def __init__(
        self,
        server="ldap",
        port=None,
        base_dn="ou=users,dc=domain,dc=com",
        mode="uid",
        secure=False,
        self_signed_certificate=None,  # See NOTE below
        cert_path=None,
        cert_file=None,
        cacert_path=None,
        cacert_file=None,
        key_file=None,
        bind_dn=None,
        bind_pw=None,
        filterstr="objectClass=*",
        username_attrib="uid",
        custom_scope="subtree",
        allowed_groups=None,
        manage_user=False,
        user_firstname_attrib="cn:1",
        user_lastname_attrib="cn:2",
        user_mail_attrib="mail",
        manage_groups=False,
        manage_groups_callback=[],
        db=None,
        group_dn=None,
        group_name_attrib="cn",
        group_member_attrib="memberUid",
        group_filterstr="objectClass=*",
        group_mapping={},
        tls=False,
        logger=logging,
        groups=None,
    ):
        self.server = server
        self.port = port
        self.base_dn = base_dn
        self.mode = mode
        self.secure = secure
        self.self_signed_certificate = self_signed_certificate
        self.cert_path = cert_path
        self.cert_file = cert_file
        self.cacert_path = cacert_path
        self.cacert_file = cacert_file
        self.key_file = key_file
        self.bind_dn = bind_dn
        self.bind_pw = bind_pw
        self.filterstr = filterstr
        self.username_attrib = username_attrib
        self.custom_scope = custom_scope
        self.allowed_groups = allowed_groups
        self.manage_user = manage_user
        self.user_firstname_attrib = user_firstname_attrib
        self.user_lastname_attrib = user_lastname_attrib
        self.user_mail_attrib = user_mail_attrib
        self.manage_groups = manage_groups
        self.manage_groups_callback = manage_groups_callback
        self.db = db
        self.group_dn = group_dn
        self.group_name_attrib = group_name_attrib
        self.group_member_attrib = group_member_attrib
        self.group_filterstr = group_filterstr
        self.group_mapping = group_mapping
        self.tls = tls
        self.logger = logger
        # rfc4515 syntax
        self.filterstr = self.filterstr.lstrip("(").rstrip(")")
        self.groups = groups

    def check_credentials(self, username, password):
        base_dn = self.base_dn
        mode = self.mode
        bind_dn = self.bind_dn
        bind_pw = self.bind_pw
        filterstr = self.filterstr
        username_attrib = self.username_attrib
        custom_scope = self.custom_scope
        manage_user = self.manage_user
        user_firstname_attrib = self.user_firstname_attrib
        user_lastname_attrib = self.user_lastname_attrib
        user_mail_attrib = self.user_mail_attrib
        manage_groups = self.manage_groups
        db = self.db
        group_mapping = self.group_mapping
        logger = self.logger

        if password == "":  # http://tools.ietf.org/html/rfc4513#section-5.1.2
            logger.warning("blank password not allowed")
            return False
        logger.debug(
            f"mode: {str(mode)}, manage_user: {str(manage_user)}, \
            custom_scope: {str(custom_scope)}, manage_groups: {str(manage_groups)}"
        )
        if manage_user:
            if user_firstname_attrib.count(":") > 0:
                (
                    user_firstname_attrib,
                    user_firstname_part,
                ) = user_firstname_attrib.split(":", 1)
                user_firstname_part = int(user_firstname_part) - 1
            else:
                user_firstname_part = None
            if user_lastname_attrib.count(":") > 0:
                (user_lastname_attrib, user_lastname_part) = user_lastname_attrib.split(
                    ":", 1
                )
                user_lastname_part = int(user_lastname_part) - 1
            else:
                user_lastname_part = None
            user_firstname_attrib = ldap.filter.escape_filter_chars(
                user_firstname_attrib
            )
            user_lastname_attrib = ldap.filter.escape_filter_chars(user_lastname_attrib)
            user_mail_attrib = ldap.filter.escape_filter_chars(user_mail_attrib)
        try:
            # if allowed_groups:
            #    if not self.is_user_in_allowed_groups(
            #        username, password, allowed_groups
            #    ):
            #        return False
            con = self._init_ldap()
            if mode == "ad":
                # Microsoft Active Directory
                if "@" not in username:
                    domain = []
                    for x in base_dn.split(","):
                        if "DC=" in x.upper():
                            domain.append(x.split("=")[-1])
                    username = f"{username}@{'.'.join(domain)}"
                username_bare = username.split("@")[0]
                con.set_option(ldap.OPT_PROTOCOL_VERSION, 3)
                # In cases where ForestDnsZones and DomainDnsZones are found,
                # result will look like the following:
                # ['ldap://ForestDnsZones.domain.com/DC=ForestDnsZones,
                #    DC=domain,DC=com']
                if bind_dn:
                    # need to search directory with an admin account 1st
                    con.simple_bind_s(bind_dn, bind_pw)
                else:
                    # credentials should be in the form of username@domain.tld
                    con.simple_bind_s(username, password)
                # this will throw an index error if the account is not found
                # in the base_dn
                requested_attrs = ["sAMAccountName"]
                if manage_user:
                    requested_attrs.extend(
                        [user_firstname_attrib, user_lastname_attrib, user_mail_attrib]
                    )

                result = con.search_ext_s(
                    base_dn,
                    ldap.SCOPE_SUBTREE,
                    f"(&(sAMAccountName={ldap.filter.escape_filter_chars(username_bare)})({filterstr}))",
                    requested_attrs,
                )[0][1]
                logger.info(f"Login result: {result}")
                if not isinstance(result, dict):
                    # result should be a dict in the form
                    # {'sAMAccountName': [username_bare]}
                    logger.warning(f"User {username} not found!")
                    return False
                if bind_dn:
                    # We know the user exists & is in the correct OU
                    # so now we just check the password
                    con.simple_bind_s(username, password)
                username = username_bare

            if mode == "domino":
                # Notes Domino
                if "@" in username:
                    username = username.split("@")[0]
                con.simple_bind_s(username, password)
                if manage_user:
                    # TODO: sorry I have no clue how to query attrs in domino
                    result = {
                        user_firstname_attrib: username,
                        user_lastname_attrib: None,
                        user_mail_attrib: None,
                    }

            if mode == "cn":
                # OpenLDAP (CN)
                if bind_dn and bind_pw:
                    con.simple_bind_s(bind_dn, bind_pw)
                dn = "cn=" + username + "," + base_dn
                con.simple_bind_s(dn, password)
                if manage_user:
                    result = con.search_s(
                        dn,
                        ldap.SCOPE_BASE,
                        "(objectClass=*)",
                        [user_firstname_attrib, user_lastname_attrib, user_mail_attrib],
                    )[0][1]

            if mode == "uid":
                # OpenLDAP (UID)
                if bind_dn and bind_pw:
                    con.simple_bind_s(bind_dn, bind_pw)
                    dn = "uid=" + username + "," + base_dn
                    dn = con.search_s(
                        base_dn, ldap.SCOPE_SUBTREE, f"(uid={username, ['']})"
                    )[0][0]
                else:
                    dn = "uid=" + username + "," + base_dn
                con.simple_bind_s(dn, password)
                if manage_user:
                    result = con.search_s(
                        dn,
                        ldap.SCOPE_BASE,
                        "(objectClass=*)",
                        [user_firstname_attrib, user_lastname_attrib, user_mail_attrib],
                    )[0][1]

            if mode == "company":
                # no DNs or password needed to search directory
                dn = ""
                pw = ""
                # bind anonymously
                con.simple_bind_s(dn, pw)
                # search by e-mail address
                filter = f"(&(mail={ldap.filter.escape_filter_chars(username)})({filterstr}))"
                # find the uid
                attrs = ["uid"]
                if manage_user:
                    attrs.extend(
                        [user_firstname_attrib, user_lastname_attrib, user_mail_attrib]
                    )
                # perform the actual search
                company_search_result = con.search_s(
                    base_dn, ldap.SCOPE_SUBTREE, filter, attrs
                )
                dn = company_search_result[0][0]
                result = company_search_result[0][1]
                # perform the real authentication test
                con.simple_bind_s(dn, password)

            if mode == "uid_r":
                # OpenLDAP (UID) with subtree search and multiple DNs
                if isinstance(base_dn, list):
                    basedns = base_dn
                else:
                    basedns = [base_dn]
                filter = (
                    f"(&(uid={ldap.filter.escape_filter_chars(username)})({filterstr}))"
                )
                found = False
                for basedn in basedns:
                    try:
                        result = con.search_s(basedn, ldap.SCOPE_SUBTREE, filter)
                        if result:
                            user_dn = result[0][0]
                            # Check the password
                            con.simple_bind_s(user_dn, password)
                            found = True
                            break
                    except ldap.LDAPError:
                        (exc_type, exc_value) = sys.exc_info()[:2]
                        logger.warning(
                            f"ldap_auth: searching {basedn} for {filter} resulted in {exc_type}: {exc_value}\n"
                        )
                if not found:
                    logger.warning(f"User [{username}] not found!")
                    return False
                result = result[0][1]
            if mode == "custom":
                # OpenLDAP (username_attrs) with subtree search and
                # multiple DNs
                if isinstance(base_dn, list):
                    basedns = base_dn
                else:
                    basedns = [base_dn]
                filter = f"(&({username_attrib}={ldap.filter.escape_filter_chars(username)})({filterstr}))"
                if custom_scope == "subtree":
                    scope = ldap.SCOPE_SUBTREE
                elif custom_scope == "base":
                    scope = ldap.SCOPE_BASE
                elif custom_scope == "onelevel":
                    scope = ldap.SCOPE_ONELEVEL
                found = False
                for basedn in basedns:
                    try:
                        result = con.search_s(basedn, scope, filter)
                        if result:
                            user_dn = result[0][0]
                            # Check the password
                            con.simple_bind_s(user_dn, password)
                            found = True
                            break
                    except ldap.LDAPError:
                        (exc_type, exc_value) = sys.exc_info()[:2]
                        logger.warning(
                            f"ldap_auth: searching {basedn} for {filter} resulted in {exc_type}: {exc_value}\n"
                        )
                if not found:
                    logger.warning(f"User {username} not found!")
                    return False
                result = result[0][1]
            if manage_user:
                logger.info(f"[{str(username)}] Manage user data")
                try:
                    store_sso_id = "ldap:" + username
                    user_firstname = result[user_firstname_attrib][0]
                    if user_firstname_part is not None:
                        store_user_firstname = user_firstname.split(
                            b" " if isinstance(user_firstname, bytes) else " ", 1
                        )[user_firstname_part]
                    else:
                        store_user_firstname = user_firstname
                except KeyError:
                    store_user_firstname = None
                try:
                    user_lastname = result[user_lastname_attrib][0]
                    if user_lastname_part is not None:
                        store_user_lastname = user_lastname.split(
                            b" " if isinstance(user_lastname, bytes) else " ", 1
                        )[user_lastname_part]
                    else:
                        store_user_lastname = user_lastname
                except KeyError:
                    store_user_lastname = None
                try:
                    store_user_mail = result[user_mail_attrib][0]
                except KeyError:
                    store_user_mail = None
                update_or_insert_values = {
                    "first_name": store_user_firstname,
                    "last_name": store_user_lastname,
                    "email": store_user_mail,
                    "username": username,
                    "sso_id": store_sso_id,
                }
                if "@" not in username:
                    # user as username
                    # ################
                    fields = ["first_name", "last_name", "email"]
                    user_in_db = db(db.auth_user.username == username)
                elif "@" in username:
                    # user as email
                    # #############
                    fields = ["first_name", "last_name"]
                    user_in_db = db(db.auth_user.email == username)
                    update_or_insert_values = dict(
                        ((f, update_or_insert_values[f]) for f in fields)
                    )

                if user_in_db.count() > 0:
                    actual_values = (
                        user_in_db.select(*[db.auth_user[f] for f in fields])
                        .first()
                        .as_dict()
                    )
                    if (
                        update_or_insert_values != actual_values
                    ):  # We don't update record if values are the same
                        user_in_db.update(**update_or_insert_values)
                else:
                    db.auth_user.insert(**update_or_insert_values)

                if manage_groups:
                    if not self.do_manage_groups(con, username, group_mapping):
                        return False

            con.unbind()
            return True
        except ldap.INVALID_CREDENTIALS:
            return False
        except ldap.LDAPError:
            import traceback

            logger.warning(f"[{str(username)}] Error in Ldap processing")
            logger.debug(traceback.format_exc())
            print(traceback.format_exc())
            return False
        except IndexError:  # for AD membership test
            import traceback

            logger.warning(f"[{str(username)}] Ldap result indexing error")
            logger.debug(traceback.format_exc())
            return False

    def is_user_in_allowed_groups(self, username, password=None):
        allowed_groups = self.allowed_groups

        """
        Figure out if the username is a member of an allowed group
        in ldap or not
        """
        #
        # Get all group name where the user is in actually in ldap
        # #########################################################
        self.groups = self.get_user_groups_from_ldap(username, password)

        # search for allowed group names
        if not isinstance(allowed_groups, list):
            allowed_groups = [allowed_groups]
        for group in allowed_groups:
            if group in self.groups:
                # Match
                return True
        # No match
        return False

    def do_manage_groups(self, con, username, group_mapping={}):
        """
        Manage user groups

        Get all user's groups from ldap and refresh the already stored
        ones in py4web's application database or create new groups
        according to ldap.
        """
        manage_groups_callback = self.manage_groups_callback
        db = self.db
        group_mapping = self.group_mapping
        logger = self.logger
        groups = self.groups

        logger.info(f"[{str(username)}] Manage user groups")
        try:
            #
            # Get all group name where the user is in actually in ldap
            # #########################################################
            ldap_groups_of_the_user = self.get_user_groups_from_ldap(con, username)

            if group_mapping != {}:
                l = []
                for group in ldap_groups_of_the_user:
                    if group in group_mapping:
                        l.append(group_mapping[group])
                ldap_groups_of_the_user = l
                logger.info(f"User groups after remapping: {str(l)}")

            #
            # Get all group name where the user is in actually in local db
            # #############################################################
            try:
                db_user_id = (
                    db(db.auth_user.username == username)
                    .select(db.auth_user.id)
                    .first()
                    .id
                )
            except:
                try:
                    db_user_id = (
                        db(db.auth_user.email == username)
                        .select(db.auth_user.id)
                        .first()
                        .id
                    )
                except AttributeError:
                    #
                    # There is no user in local db
                    # We create one
                    # ##############################
                    try:
                        db_user_id = db.auth_user.insert(
                            username=username, first_name=username
                        )
                    except AttributeError:
                        db_user_id = db.auth_user.insert(
                            email=username, first_name=username
                        )
            if not db_user_id:
                logger.error(f"There is no username or email for {username}!")
                raise
            db_groups_of_the_user = groups.get(db_user_id)

            auth_membership_changed = False
            #
            # Delete user membership from groups where user is not anymore
            # #############################################################
            for group_to_del in db_groups_of_the_user:
                if ldap_groups_of_the_user.count(group_to_del) == 0:
                    groups.remove(db_user_id, group_to_del)
                    auth_membership_changed = True

            #
            # Create user membership in groups where user is not in already
            # ##############################################################
            for group_to_add in ldap_groups_of_the_user:
                if db_groups_of_the_user.count(group_to_add) == 0:
                    groups.add(db_user_id, group_to_add)
                    auth_membership_changed = True

            if auth_membership_changed:
                for callback in manage_groups_callback:
                    callback()
        except:
            logger.warning(f"[{str(username)}] Groups are not managed successfully!")
            import traceback

            logger.debug(traceback.format_exc())
            return False
        return True

    def _init_ldap(self):
        """
        Inizialize ldap connection
        """

        server = self.server
        port = self.port
        secure = self.secure
        self_signed_certificate = self.self_signed_certificate
        cert_file = self.cert_file
        cacert_path = self.cacert_path
        cacert_file = self.cacert_file
        key_file = self.key_file
        tls = self.tls
        logger = self.logger

        logger.debug(f"[{str(server)}] Initialize ldap connection")
        if secure:
            if not port:
                port = 636

            logger.debug(f"Using Secure LDAP connection to {server}:{port}")
            if self_signed_certificate:
                # NOTE : If you have a self-signed SSL Certificate pointing over "port=686" and "secure=True" alone
                #        will not work, you need also to set "self_signed_certificate=True".
                # Ref1: https://onemoretech.wordpress.com/2015/06/25/connecting-to-ldap-over-self-signed-tls-with-python/
                # Ref2: http://bneijt.nl/blog/post/connecting-to-ldaps-with-self-signed-cert-using-python/
                ldap.set_option(ldap.OPT_X_TLS_REQUIRE_CERT, ldap.OPT_X_TLS_NEVER)

            if cacert_path:
                ldap.set_option(ldap.OPT_X_TLS_CACERTDIR, cacert_path)

            if cacert_file:
                ldap.set_option(ldap.OPT_X_TLS_REQUIRE_CERT, ldap.OPT_X_TLS_NEVER)
                ldap.set_option(ldap.OPT_X_TLS_CACERTFILE, cacert_file)
            if cert_file:
                ldap.set_option(ldap.OPT_X_TLS_CERTFILE, cert_file)
            if key_file:
                ldap.set_option(ldap.OPT_X_TLS_KEYFILE, key_file)

            con = ldap.initialize("ldaps://" + server + ":" + str(port))
        else:
            if not port:
                port = 389
            logger.debug(f"Initializing LDAP connection to ldap://{server}:{str(port)}")
            con = ldap.initialize(f"ldap://{server}:{str(port)}")
        if tls:
            con.start_tls_s()
        return con

    def get_user_groups_from_ldap(self, username=None, password=None):
        """
        Get all group names from ldap where the user is in
        """
        bind_dn = self.bind_dn
        bind_pw = self.bind_pw
        base_dn = self.base_dn
        mode = self.mode
        filterstr = self.filterstr
        group_dn = self.group_dn
        group_name_attrib = self.group_name_attrib
        group_member_attrib = self.group_member_attrib
        group_filterstr = self.group_filterstr
        logger = self.logger

        # if username is None, return empty list
        if username is None:
            return []

        logger.info(f"[{str(username)}] Get user groups from ldap")
        #
        # Get all group name where the user is in actually in ldap
        # #########################################################
        if mode == "ad":
            #
            # Get the AD username
            # ####################
            if "@" not in username:
                domain = []
                for x in base_dn.split(","):
                    if "DC=" in x.upper():
                        domain.append(x.split("=")[-1])
                username = f"{username}@{'.'.join(domain)}"
            username_bare = username.split("@")[0]
            con = self._init_ldap()
            con.set_option(ldap.OPT_PROTOCOL_VERSION, 3)
            try:
                if bind_dn:
                    # need to search directory with an admin account 1st
                    con.simple_bind_s(bind_dn, bind_pw)
                else:
                    # credentials should be in the form of user...@domain.tld
                    con.simple_bind_s(username, password)
            except (ldap.INVALID_CREDENTIALS, ldap.OPERATIONS_ERROR):
                return []
            # In cases where ForestDnsZones and DomainDnsZones are found,
            # result will look like the following:
            # ['ldap://ForestDnsZones.domain.com/DC=ForestDnsZones,
            #     DC=domain,DC=com']
            # We have to use the full string
            bare = ldap.filter.escape_filter_chars(username_bare)
            username = con.search_ext_s(
                base_dn,
                ldap.SCOPE_SUBTREE,
                f"(&(sAMAccountName={bare})({filterstr}))",
                ["cn"],
            )[0][0]

        # if username is None, return empty list
        if username is None:
            return []
        # search for groups where user is in
        filter = f"(&({ldap.filter.escape_filter_chars(group_member_attrib)}=\
            {ldap.filter.escape_filter_chars(username)})({group_filterstr}))"

        group_search_result = con.search_s(
            group_dn, ldap.SCOPE_SUBTREE, filter, [group_name_attrib]
        )
        ldap_groups_of_the_user = list()
        for group_row in group_search_result:
            group = group_row[1]
            if isinstance(group, dict) and group_name_attrib in group:
                ldap_groups_of_the_user.append(
                    str(group[group_name_attrib][0], encoding="utf-8")
                )

        logger.debug(f"User groups for {username}: {ldap_groups_of_the_user}")
        return list(ldap_groups_of_the_user)



# from web2py

from __future__ import print_function

import datetime
import random
import re
import sys
import uuid

from pydal._compat import pickle, unicodeT

IUP = {
    "ad": {"minim": 1},
    "irure": {"dolor": 1},
    "in": {"voluptate": 1, "reprehenderit": 1, "culpa": 1},
    "ea": {"commodo": 1},
    "excepteur": {"sint": 1},
    "sunt": {"in": 1},
    "elit": {"sed": 1},
    "duis": {"aute": 1},
    "sed": {"do": 1},
    "eiusmod": {"tempor": 1},
    "enim": {"ad": 1},
    "eu": {"fugiat": 1},
    "et": {"dolore": 1},
    "labore": {"et": 1},
    "incididunt": {"ut": 1},
    "reprehenderit": {"in": 1},
    "est": {"laborum": 1},
    "quis": {"nostrud": 1},
    "sit": {"amet": 1},
    "deserunt": {"mollit": 1},
    "nostrud": {"exercitation": 1},
    "qui": {"officia": 1},
    ".": {"excepteur": 1, "ut": 1, "duis": 1},
    "consectetur": {"adipiscing": 1},
    "aute": {"irure": 1},
    "dolore": {"eu": 1, "magna": 1},
    "mollit": {"anim": 1},
    "aliquip": {"ex": 1},
    "nulla": {"pariatur": 1},
    "laborum": {"": 1},
    "do": {"eiusmod": 1},
    "non": {"proident": 1},
    "commodo": {"consequat": 1},
    "aliqua": {".": 1},
    "cillum": {"dolore": 1},
    "sint": {"occaecat": 1},
    "velit": {"esse": 1},
    "officia": {"deserunt": 1},
    "veniam": {"quis": 1},
    "consequat": {".": 1},
    "magna": {"aliqua": 1},
    "cupidatat": {"non": 1},
    "ullamco": {"laboris": 1},
    "lorem": {"ipsum": 1},
    "amet": {"consectetur": 1},
    "ipsum": {"dolor": 1},
    "nisi": {"ut": 1},
    "fugiat": {"nulla": 1},
    "occaecat": {"cupidatat": 1},
    "proident": {"sunt": 1},
    "ut": {"aliquip": 1, "labore": 1, "enim": 1},
    "minim": {"veniam": 1},
    "culpa": {"qui": 1},
    "tempor": {"incididunt": 1},
    "pariatur": {".": 1},
    "laboris": {"nisi": 1},
    "anim": {"id": 1},
    "adipiscing": {"elit": 1},
    "id": {"est": 1},
    "dolor": {"in": 1, "sit": 1},
    "voluptate": {"velit": 1},
    "esse": {"cillum": 1},
    "exercitation": {"ullamco": 1},
    "ex": {"ea": 1},
}

FIRST_NAMES = "James,John,Robert,Michael,William,David,Richard,Charles,Joseph,Thomas,Christopher,Daniel,Paul,Mark,Donald,George,Kenneth,Steven,Edward,Brian,Ronald,Anthony,Kevin,Jason,Matthew,Gary,Timothy,Jose,Larry,Jeffrey,Frank,Scott,Eric,Stephen,Andrew,Raymond,Gregory,Joshua,Jerry,Dennis,Walter,Patrick,Peter,Harold,Douglas,Henry,Carl,Arthur,Ryan,Roger,Joe,Juan,Jack,Albert,Jonathan,Justin,Terry,Gerald,Keith,Samuel,Willie,Ralph,Lawrence,Nicholas,Roy,Benjamin,Bruce,Brandon,Adam,Harry,Fred,Wayne,Billy,Steve,Louis,Jeremy,Aaron,Randy,Howard,Eugene,Carlos,Russell,Bobby,Victor,Martin,Ernest,Phillip,Todd,Jesse,Craig,Alan,Shawn,Clarence,Sean,Philip,Chris,Johnny,Earl,Jimmy,Antonio,Danny,Bryan,Tony,Luis,Mike,Stanley,Leonard,Nathan,Dale,Manuel,Rodney,Curtis,Norman,Allen,Marvin,Vincent,Glenn,Jeffery,Travis,Jeff,Chad,Jacob,Lee,Melvin,Alfred,Kyle,Francis,Bradley,Jesus,Herbert,Frederick,Ray,Joel,Edwin,Don,Eddie,Ricky,Troy,Randall,Barry,Alexander,Bernard,Mario,Leroy,Francisco,Marcus,Micheal,Theodore,Clifford,Miguel,Oscar,Jay,Jim,Tom,Calvin,Alex,Jon,Ronnie,Bill,Lloyd,Tommy,Leon,Derek,Warren,Darrell,Jerome,Floyd,Leo,Alvin,Tim,Wesley,Gordon,Dean,Greg,Jorge,Dustin,Pedro,Derrick,Dan,Lewis,Zachary,Corey,Herman,Maurice,Vernon,Roberto,Clyde,Glen,Hector,Shane,Ricardo,Sam,Rick,Lester,Brent,Ramon,Charlie,Tyler,Gilbert,Gene,Marc,Reginald,Ruben,Brett,Angel,Nathaniel,Rafael,Leslie,Edgar,Milton,Raul,Ben,Chester,Cecil,Duane,Franklin,Andre,Elmer,Brad,Gabriel,Ron,Mitchell,Roland,Arnold,Harvey,Jared,Adrian,Karl,Cory,Claude,Erik,Darryl,Jamie,Neil,Jessie,Christian,Javier,Fernando,Clinton,Ted,Mathew,Tyrone,Darren,Lonnie,Lance,Cody,Julio,Kelly,Kurt,Allan,Nelson,Guy,Clayton,Hugh,Max,Dwayne,Dwight,Armando,Felix,Jimmie,Everett,Jordan,Ian,Wallace,Ken,Bob,Jaime,Casey,Alfredo,Alberto,Dave,Ivan,Johnnie,Sidney,Byron,Julian,Isaac,Morris,Clifton,Willard,Daryl,Ross,Virgil,Andy,Marshall,Salvador,Perry,Kirk,Sergio,Marion,Tracy,Seth,Kent,Terrance,Rene,Eduardo,Terrence,Enrique,Freddie,Wade,Austin,Stuart,Fredrick,Arturo,Alejandro,Jackie,Joey,Nick,Luther,Wendell,Jeremiah,Evan,Julius,Dana,Donnie,Otis,Shannon,Trevor,Oliver,Luke,Homer,Gerard,Doug,Kenny,Hubert,Angelo".split(
    ","
)

LAST_NAMES = "Smith,Johnson,Williams,Jones,Brown,Davis,Miller,Wilson,Moore,Taylor,Anderson,Thomas,Jackson,White,Harris,Martin,Thompson,Garcia,Martinez,Robinson,Clark,Rodriguez,Lewis,Lee,Walker,Hall,Allen,Young,Hernandez,King,Wright,Lopez,Hill,Scott,Green,Adams,Baker,Gonzalez,Nelson,Carter,Mitchell,Perez,Roberts,Turner,Phillips,Campbell,Parker,Evans,Edwards,Collins,Stewart,Sanchez,Morris,Rogers,Reed,Cook,Morgan,Bell,Murphy,Bailey,Rivera,Cooper,Richardson,Cox,Howard,Ward,Torres,Peterson,Gray,Ramirez,James,Watson,Brooks,Kelly,Sanders,Price,Bennett,Wood,Barnes,Ross,Henderson,Coleman,Jenkins,Perry,Powell,Long,Patterson,Hughes,Flores,Washington,Butler,Simmons,Foster,Gonzales,Bryant,Alexander,Russell,Griffin,Diaz,Hayes,Myers,Ford,Hamilton,Graham,Sullivan,Wallace,Woods,Cole,West,Jordan,Owens,Reynolds,Fisher,Ellis,Harrison,Gibson,Mcdonald,Cruz,Marshall,Ortiz,Gomez,Murray,Freeman,Wells,Webb,Simpson,Stevens,Tucker,Porter,Hunter,Hicks,Crawford,Henry,Boyd,Mason,Morales,Kennedy,Warren,Dixon,Ramos,Reyes,Burns,Gordon,Shaw,Holmes,Rice,Robertson,Hunt,Black,Daniels,Palmer,Mills,Nichols,Grant,Knight,Ferguson,Rose,Stone,Hawkins,Dunn,Perkins,Hudson,Spencer,Gardner,Stephens,Payne,Pierce,Berry,Matthews,Arnold,Wagner,Willis,Ray,Watkins,Olson,Carroll,Duncan,Snyder,Hart,Cunningham,Bradley,Lane,Andrews,Ruiz,Harper,Fox,Riley,Armstrong,Carpenter,Weaver,Greene,Lawrence,Elliott,Chavez,Sims,Austin,Peters,Kelley,Franklin,Lawson,Fields,Gutierrez,Ryan,Schmidt,Carr,Vasquez,Castillo,Wheeler,Chapman,Oliver,Montgomery,Richards,Williamson,Johnston,Banks,Meyer,Bishop,Mccoy,Howell,Alvarez,Morrison,Hansen,Fernandez,Garza,Harvey,Little,Burton,Stanley,Nguyen,George,Jacobs,Reid,Kim,Fuller,Lynch,Dean,Gilbert,Garrett,Romero,Welch,Larson,Frazier,Burke,Hanson,Day,Mendoza,Moreno,Bowman,Medina,Fowler,Brewer,Hoffman,Carlson,Silva,Pearson,Holland,Douglas,Fleming,Jensen,Vargas,Byrd,Davidson,Hopkins,May,Terry,Herrera,Wade,Soto,Walters".split(
    ","
)


class Learner:
    def __init__(self):
        self.db = {}

    def learn(self, text):
        replacements1 = {
            r"[^a-zA-Z0-9\.;:\-]": " ",
            r"\s+": " ",
            ", ": " , ",
            r"\. ": " . ",
            ": ": " : ",
            "; ": " ; ",
        }
        for key, value in replacements1.items():
            text = re.sub(key, value, text)
        items = [item.lower() for item in text.split(" ")]
        for i in range(len(items) - 1):
            item = items[i]
            nextitem = items[i + 1]
            if item not in self.db:
                self.db[item] = {}
            if nextitem not in self.db[item]:
                self.db[item][nextitem] = 1
            else:
                self.db[item][nextitem] += 1

    def save(self, filename):
        with open(filename, "wb") as fp:
            pickle.dump(self.db, fp)

    def load(self, filename):
        with open(filename, "rb") as fp:
            self.loadd(pickle.load(fp))

    def loadd(self, db):
        self.db = db

    def generate(self, length=10000, prefix=False):
        replacements2 = {" ,": ",", " \\.": ".\n", " :": ":", " ;": ";", "\n\\s+": "\n"}
        keys = list(self.db.keys())
        key = keys[random.randint(0, len(keys) - 1)]
        words = key
        words = words.capitalize()
        regex = re.compile("[a-z]+")
        for i in range(length):
            okey = key
            if key not in self.db:
                break  # should not happen
            db = self.db[key]
            s = sum(db.values())
            i = random.randint(0, s - 1)
            for key, value in db.items():
                if i < value:
                    break
                else:
                    i -= value
            if okey == ".":
                key1 = key.capitalize()
            else:
                key1 = key
            if prefix and regex.findall(key1) and random.random() < 0.01:
                key1 = '<a href="%s%s">%s</a>' % (prefix, key1, key1)
            words += " " + key1
        text = words
        for key, value in replacements2.items():
            text = re.sub(key, value, text)
        return text + ".\n"


def da_du_ma(n=4):
    return "".join(
        [
            ["da", "du", "ma", "mo", "ce", "co", "pa", "po", "sa", "so", "ta", "to"][
                random.randint(0, 11)
            ]
            for i in range(n)
        ]
    )


def populate(table, n=None, default=False, compute=False, contents=None, ell=None):
    """Populate table with n records.

    if n is None, it does not populate the database but returns a generator
    if default=True use default values to fields.
    if compute=False doesn't load values into computed fields.
    if contents has data, use these values to populate related fields.

    can be used in two ways:

    >>> populate(db.tablename, n=100)

    or

    >>> for k,row in enumerate(populate(db.tablename)): print row
    """

    contents = contents or {}

    generator = populate_generator(
        table, default=default, compute=compute, contents=contents, ell=None
    )
    if n is not None:
        for k, record in enumerate(generator):
            if k >= n:
                break
            table.insert(**record)
        table._db.commit()

    return generator


def populate_generator(table, default=True, compute=False, contents=None, ell=None):
    """Populate table with n records.

    if default=True use default values to fields.
    if compute=False doesn't load values into computed fields.
    if contents has data, use these values to populate related fields.
    """
    contents = contents or {}

    if not ell:
        ell = Learner()
        ell.loadd(IUP)

    db = table._db
    ids = {}

    while True:
        record = contents.copy()  # load user supplied contents.
        first_name = random.choice(FIRST_NAMES)
        username = first_name.lower() + str(random.randint(1000, 9999))
        for fieldname in table.fields:
            if fieldname in record:
                continue  # if user supplied it, let it be.

            field = table[fieldname]
            if not isinstance(field.type, (str, unicodeT)):
                continue
            elif field.type == "id":
                continue
            elif field.type == "upload":
                continue
            elif field.compute is not None:
                continue
            elif default and field.default not in (None, ""):
                record[fieldname] = field.default
            elif compute and field.compute:
                continue
            elif field.type == "boolean":
                record[fieldname] = random.random() > 0.5
            elif field.type == "date":
                record[fieldname] = datetime.date(2009, 1, 1) - datetime.timedelta(
                    days=random.randint(0, 365)
                )
            elif field.type == "datetime":
                record[fieldname] = datetime.datetime(2009, 1, 1) - datetime.timedelta(
                    days=random.randint(0, 365)
                )
            elif field.type == "time":
                h = random.randint(0, 23)
                m = 15 * random.randint(0, 3)
                record[fieldname] = datetime.time(h, m, 0)
            elif field.type == "password":
                record[fieldname] = ""
            elif field.type == "integer" and hasattr(field.requires, "options"):
                options = field.requires.options(zero=False)
                if len(options) > 0:
                    record[fieldname] = options[random.randint(0, len(options) - 1)][0]
                else:
                    record[fieldname] = None
            elif field.type == "list:integer" and hasattr(field.requires, "options"):
                options = field.requires.options(zero=False)
                if len(options) > 0:
                    record[fieldname] = [
                        item[0]
                        for item in random.sample(
                            options, random.randint(0, len(options) - 1) / 2
                        )
                    ]
            elif field.type == "integer":
                try:
                    record[fieldname] = random.randint(
                        field.requires.minimum, field.requires.maximum - 1
                    )
                except Exception:
                    if "day" in fieldname:
                        record[fieldname] = random.randint(1, 28)
                    elif "month" in fieldname:
                        record[fieldname] = random.randint(1, 12)
                    elif "year" in fieldname:
                        record[fieldname] = random.randint(2000, 2013)
                    else:
                        record[fieldname] = random.randint(0, 1000)
            elif field.type in ("float", "double") or str(field.type).startswith(
                "decimal"
            ):
                if hasattr(field.requires, "minimum"):
                    rand = random.random()
                    if str(field.type).startswith("decimal"):
                        import decimal

                        rand = decimal.Decimal(rand)
                    record[fieldname] = field.requires.minimum + rand * (
                        field.requires.maximum - field.requires.minimum
                    )
                else:
                    record[fieldname] = random.random() * 1000
            elif field.type[:10] == "reference ":
                tablename = field.type[10:]
                rtable = db[tablename]
                if tablename not in ids:
                    if db._dbname == "gql":
                        ids[tablename] = [x.id for x in db(rtable).select(rtable.id)]
                    else:
                        ids[tablename] = [x.id for x in db(rtable).select(rtable.id)]
                n = len(ids[tablename])
                if n:
                    record[fieldname] = ids[tablename][random.randint(0, n - 1)]
                else:
                    record[fieldname] = 0
            elif field.type[:15] == "list:reference ":
                tablename = field.type[15:]
                rtable = db[tablename]
                if tablename not in ids:
                    if db._dbname == "gql":
                        ids[tablename] = [x.id for x in db(rtable).select(rtable.id)]
                    else:
                        ids[tablename] = [x.id for x in db(rtable).select(rtable.id)]
                n = len(ids[tablename])
                if n:
                    record[fieldname] = [
                        item
                        for item in random.sample(
                            ids[tablename], random.randint(0, n - 1) / 2
                        )
                    ]
                else:
                    record[fieldname] = 0
            elif field.type == "list:string" and hasattr(field.requires, "options"):
                options = field.requires.options(zero=False)
                if len(options) > 0:
                    record[fieldname] = [
                        item[0]
                        for item in random.sample(
                            options, random.randint(0, len(options) - 1) / 2
                        )
                    ]
            elif field.type == "string":
                if hasattr(field.requires, "options"):
                    options = field.requires.options(zero=False)
                    record[fieldname] = options[random.randint(0, len(options) - 1)][0]
                elif fieldname.find("url") >= 0:
                    record[fieldname] = "http://%s.example.com" % str(uuid.uuid4())[:8]
                elif fieldname.find("email") >= 0:
                    record[fieldname] = "%s@example.com" % username
                elif fieldname.find("name") >= 0:
                    if fieldname.find("first") >= 0:
                        record[fieldname] = first_name
                    elif fieldname.find("last") >= 0:
                        record[fieldname] = random.choice(LAST_NAMES)
                    elif fieldname.find("username") >= 0:
                        record[fieldname] = username
                    else:
                        record[fieldname] = (
                            random.choice(FIRST_NAMES) + " " + random.choice(LAST_NAMES)
                        )
                elif fieldname.find("phone") >= 0:
                    record[fieldname] = "(%s%s%s) %s%s%s-%s%s%s%s" % (
                        random.choice("1234567890"),
                        random.choice("1234567890"),
                        random.choice("1234567890"),
                        random.choice("1234567890"),
                        random.choice("1234567890"),
                        random.choice("1234567890"),
                        random.choice("1234567890"),
                        random.choice("1234567890"),
                        random.choice("1234567890"),
                        random.choice("1234567890"),
                    )
                elif fieldname.find("address") >= 0:
                    record[fieldname] = "%s %s %s Street" % (
                        random.randint(1000, 9000),
                        random.choice(FIRST_NAMES),
                        random.choice(LAST_NAMES),
                    )
                else:
                    z = ell.generate(10, prefix=False)
                    record[fieldname] = z[: min(60, field.length)].replace("\n", " ")
            elif field.type == "text":
                if fieldname.find("address") >= 0:
                    record[fieldname] = "%s %s %s Street\nChicago, IL\nUSA" % (
                        random.randint(1000, 9000),
                        random.choice(FIRST_NAMES),
                        random.choice(LAST_NAMES),
                    )
                else:
                    record[fieldname] = ell.generate(
                        random.randint(10, 100), prefix=None
                    )
        yield record


if __name__ == "__main__":
    ell = Learner()
    ell.db = IUP
    if len(sys.argv) > 1:
        ell.learn(open(sys.argv[1]).read())
        print(ell.db)
    print(ell.generate(100))



import sys

__author__ = "Massimo Di Pierro <massimo.dipierro@gmail.com>"
__license__ = "BSD-3-Clause"
__version__ = "1.20251012.1"


def _maybe_gevent():
    for arg in sys.argv[1:]:
        if "gevent" in arg.lower():
            from gevent import monkey

            monkey.patch_all()
            break


_maybe_gevent()


from .core import HTTP  # checks for version compatibility; bottle
from .core import URL  # custom helper
from .core import Field  # pydal
from .core import Translator  # from pluralize
from .core import action  # main py4web decorator
from .core import render  # yatl
from .core import (
    DAL,
    Cache,
    Condition,
    Flash,
    Session,
    abort,
    check_compatible,
    redirect,
    request,
    response,
    safely,
)



from . import OAuth2


class OAuth2Discord(OAuth2):
    """
    This plugin is designed to allow usage of Discord as an OAuth2 provider.

    Example usage to register plugin:

    auth.register_plugin(
        OAuth2Discord(
            client_id=<DISCORD_CLIENT_ID>,
            client_secret=<DISCORD_CLIENT_SECRET>,
            callback_url="auth/plugin/oauth2discord/callback",
        )
    )

    To obtain a Discord client ID and secret, create an application at https://discord.com/developers/applications.
    You will also have to register your OAuth2 redirect URI in your created application:
    http(s)://<your host>/<your app name>/auth/plugin/oauth2discord/callback

    Note that as Discord users have no concept of first/last name, the user in the auth table will contain the
    Discord username as the first name and discriminator as the last name.
    """

    name = "oauth2discord"
    label = "Discord"

    login_url = "https://discord.com/api/oauth2/authorize"
    token_url = "https://discord.com/api/oauth2/token"
    userinfo_url = "https://discord.com/api/v8/users/@me"
    revoke_url = "https://discord.com/api/oauth2/token/revoke"
    default_scope = "identify email"
    maps = {
        "email": "email",
        "sso_id": "id",
        "first_name": "username",
        "last_name": "discriminator",
        "username": "full_username",
    }

    def callback(self, query):
        # since Discord returns username and discriminator separately, we extend the auth plugin's callback
        # function to merge them into a single field, so it can be mapped into username correctly later
        data = super().callback(query)
        data["full_username"] = f"{data['username']}#{data['discriminator']}"
        return data



class Param:
    """
    An interface to expore predefined parameters
    which can be overwritten but not deleted or added
    Example:

    p = Param(a=1, b=2)
    p.a = 3
    p.b = 4
    print(p.a, p.b)
    p.c = 5 # Error
    """

    def __init__(self, **attr):
        self.__dict__.update(attr)

    def __setattr__(self, key, value):
        getattr(self, key)
        self.__dict__[key] = value

    def __getitem__(self, key):
        return getattr(self, key)

    __setitem__ = __setattr__



import logging
import os
import sys


def make_logger(name, loggers_info):
    """
    Abstraction layer on logging. Example of usage:

    from py4web.server_adapters.logging_utils import make_logger

    loggers_info = [
        "warning:warning.log",
        "info:info.log",
        "debug:debug.log:$(asctime)s > %(levelname)s > %(message)s",
    ]

    logger = make_logger("py4web:appname", loggers_info)
    """
    default_formatter = (
        "%(asctime)s - %(levelname)s - %(filename)s:%(lineno)d - %(message)s"
    )
    # reset loggers
    root = logging.getLogger(name)
    list(map(root.removeHandler, root.handlers))
    list(map(root.removeFilter, root.filters))
    min_level = logging.CRITICAL
    for logger in loggers_info:
        logger += ":stderr" if logger.count(":") == 0 else ""
        logger += ":" if logger.count(":") == 1 else ""
        level, filename, formatter = logger.split(":", 2)
        if not formatter:
            formatter = default_formatter
        handler_level = getattr(logging, level.upper(), logging.DEBUG)
        if handler_level < min_level:
            min_level = handler_level
        if filename in ("stdout", "stderr"):
            handler = logging.StreamHandler(getattr(sys, filename))
        else:
            handler = logging.FileHandler(filename)
        handler.setFormatter(logging.Formatter(formatter))
        handler.setLevel(handler_level)
        root.addHandler(handler)
    # Set the logger's level to the lowest handler level found.
    # This ensures that all messages at or above this level are processed by the logger,
    # and not filtered out before reaching the handlers.
    root.setLevel(min_level)
    root.propagate = False  # Prevent double logging with same name to root logger
    return root


def get_log_file(out_banner=True):
    """
    Returns the filename for logging or None
    Assumes:
    export PY4WEB_ERRORLOG=/tmp # export PY4WEB_ERRORLOG=
    if PY4WEB_ERRORLOG is :stderr or :stdout returns None
    if PY4WEB_ERRORLOG is a folder returns the name of a logfile in that dir
    if PY4WEB_ERRORLOG is a filename it returns that filename
    if the out_banner argument is true, it outputs the filename
    """
    log_dir = os.environ.get("PY4WEB_ERRORLOG", None)
    if log_dir and not log_dir.startswith(":"):
        if os.path.isdir(log_dir):
            log_file = os.path.join(log_dir, "server-py4web.log")
        else:
            log_file = log_dir
        if out_banner:
            print(f"log_file: {log_file}")
        return log_file
    return None


def check_level(level):
    "Check the level is a valid loglevel"
    # lib/python3.7/logging/__init__.py
    # CRITICAL = 50
    # FATAL = CRITICAL
    # ERROR = 40
    # WARNING = 30
    # WARN = WARNING
    # INFO = 20
    # DEBUG = 10
    # NOTSET = 0

    return (
        level
        if level
        in (
            logging.CRITICAL,
            logging.ERROR,
            logging.WARN,
            logging.INFO,
            logging.DEBUG,
            logging.NOTSET,
        )
        else logging.WARN
    )


def logging_conf(level=logging.WARN, logger_name=__name__, fmode="w"):
    "Configures logging"

    log_file = get_log_file()
    log_to = {}

    if log_file:
        if sys.version_info >= (3, 9):
            log_to["filename"] = log_file
            log_to["filemode"] = fmode
            log_to["encoding"] = "utf-8"
        else:
            h = logging.FileHandler(log_file, mode=fmode, encoding="utf-8")
            log_to.update({"handlers": [h]})

    short_msg = "%(message)s > %(threadName)s > %(asctime)s.%(msecs)03d"
    # long_msg = short_msg + " > %(funcName)s > %(filename)s:%(lineno)d > %(levelname)s"

    time_msg = "%H:%M:%S"
    # date_time_msg = '%Y-%m-%d %H:%M:%S'

    logging.basicConfig(
        format=short_msg,
        datefmt=time_msg,
        level=check_level(level),
        **log_to,
    )

    if logger_name is None:
        return None

    log = logging.getLogger("SA:" + logger_name)
    log.propagate = True

    return log



import os
import signal
import tempfile
import unittest

from click.testing import CliRunner

from py4web.core import cli


def run_cli():
    dirpath = tempfile.mkdtemp()
    dir = os.path.join(dirpath, "apps")
    runner = CliRunner()

    testargs = ["setup", dir]
    res = runner.invoke(cli, testargs, input="y\ny\ny\ny\ny\ny\ny\npassword")
    if res.exception:
        raise res.exception

    testargs = ["run", "-d", "demo", dir]
    res = runner.invoke(cli, testargs)
    if res.exception:
        raise res.exception


class MainTest(unittest.TestCase):
    def test_main(self):
        def handler(signum, frame):
            raise KeyboardInterrupt

        signal.signal(signal.SIGALRM, handler)
        signal.alarm(10)
        try:
            run_cli()
        except KeyboardInterrupt:
            pass



# -*- coding: utf-8 -*-

"""
This plugin has been developed for login through a third party application that
uses 'WordPress OAuth Server' plugin:
https://wordpress.org/plugins/miniorange-oauth-20-server
"""

from urllib.parse import urljoin

import jwt
import requests

from ...core import HTTP, URL, redirect
from . import OAuth2


class OAuth2WPMiniorange(OAuth2):
    """
    Authorized Redirect URI template to be configured in your OAuth2 server
    configuration panel:
    http(s)://<your host>/<your app name>/auth/plugin/oauth2wpminiorange/callback

    Usage example for registering plugin in your py4web application:

    auth.register_plugin(
        OAuth2WPMiniorange(
            client_id = '<your client id>',
            client_secret = '<your client secret>',
            server_host = '<your server http(s) url>'
        )
    )

    """

    # name = "oauth2wpminiorange"
    # label = "Wordpress"

    # login_url = "https://.../wp-json/moserver/authorize"
    # token_url = "https://.../wp-json/moserver/token"
    # userinfo_url = "https://.../wp-json/moserver/resource"

    default_scope = "profile openid"
    default_maps = {
        # "email": "email",
        "sso_id": "ID",
        "username": "username",
        # "first_name": "firstname",
        # "last_name": "lastname",
    }

    def __init__(
        self,
        client_id,
        client_secret,
        server_host,
        name=None,
        label="Wordpress",
        maps=None,
        login_path="/wp-json/moserver/authorize",
        token_path="/wp-json/moserver/token",
        userinfo_path="/wp-json/moserver/resource",
        **kwargs,
    ):
        name = name or self.__class__.__name__.lower()
        super(OAuth2WPMiniorange, self).__init__(
            client_id, client_secret, f"auth/plugin/{name}/callback", **kwargs
        )
        self.name = name
        self.label = label
        self.login_url = urljoin(server_host, login_path)
        self.token_url = urljoin(server_host, token_path)
        self.userinfo_url = urljoin(server_host, userinfo_path)
        self.maps = maps or self.default_maps

    ### methods that needed to be overwritten

    def _handle_callback(self, auth, get_vars):
        try:
            data = self.callback(get_vars)
        except jwt.exceptions.InvalidSignatureError as err:
            # -# In case of invalid signature jwt library raises InvalidSignatureError
            # This maybe should be common to all OAuth2 derived classes.
            raise HTTP(401, err)
        if not data:
            raise HTTP(401)
        error = data.get("error")
        if error:
            if isinstance(error, str):
                code, msg = 401, error
            else:
                code = error.get("code", 401)
                msg = error.get("message", "Unknown error")
            raise HTTP(code, msg)
        if auth.db:
            # map returned fields into auth_user fields
            user = {}
            for key, value__ in self.maps.items():
                value_, parts = data, value__.split(".")
                for part in parts:
                    # Thi is specific of the OAuth2 server implementation
                    value = value_["userinfo"][int(part) if part.isdigit() else part]

                    user[key] = value

            # -# In case of different servers username other than id sould result
            # not unique

            sso_id = user["sso_id"]
            user["sso_id"] = "%s:%s" % (
                self.name,
                sso_id,
            )
            if "username" not in user:
                user["username"] = "%s:%s" % (
                    self.name,
                    sso_id,
                )
            else:
                user["username"] = "%s:%s" % (
                    self.name,
                    user["username"],
                )
            # store or retrieve the user
            data = auth.get_or_register_user(user)

        else:
            # WIP Allow login without DB
            if "id" not in data:
                data["id"] = data.get("username") or data.get("email")
        user_id = data.get("id")
        auth.store_user_in_session(user_id)
        redirect(URL("index"))

    # -# This method is an exact copy of the one from original class
    # without this subsequent error is raised from the jwt library (don't know why)
    # jwt.exceptions.InvalidSignatureError: Signature verification failed

    def callback(self, query):
        code = query.get("code")
        statecheck = query.get("state")
        if not code:
            return False
        passed_state = self.parameters.get("passed_state")
        if passed_state is not None and statecheck != passed_state:
            return False
        data = dict(
            code=code,
            client_id=self.parameters.get("client_id"),
            client_secret=self.parameters.get("client_secret"),
            redirect_uri=URL(
                self.parameters.get("callback_url"),
                scheme=self.parameters.get("scheme"),
            ),
            grant_type="authorization_code",
        )
        res = requests.post(self.token_url, data=data)
        output = res.json()
        token = output.get("id_token")
        if token is not None:
            # Lets not get the  user attributes via the userinfo endpoint
            # but lets take the userinfo directly extracted from the token
            # res = requests.get(self.userinfo_url, headers=headers)
            data = jwt.decode(
                token,
                algorithms=self.algorithms,
                # because of this open issue
                # https://github.com/jpadilla/pyjwt/issues/359
                options={"verify_signature": False},
            )
        else:
            # fallback to old approach if "id_token" is not in the response
            token = output.get("access_token")
            headers = {"Authorization": "Bearer %s" % token}
            res = requests.get(self.userinfo_url, headers=headers)
            data = res.json()
        return data



import unittest

from py4web import URL, request


class TestURL(unittest.TestCase):
    def test_url(self):
        request.app_name = "_default"
        self.assertEqual(URL("index"), "/index")
        request.app_name = "app"
        self.assertEqual(URL("index"), "/app/index")
        self.assertEqual(URL("a", "b", vars=dict(x=1), hash="y"), "/app/a/b?x=1#y")


if __name__ == "__main__":
    unittest.main()



from py4web.core import HTTP, Fixture, request, response


class CORS(Fixture):
    """
    Fixture helper for sharing web service avoiding cross origin resource sharing problems
    """

    def __init__(self, age=86400, origin="*", headers="*", methods="*"):
        Fixture.__init__(self)

        self.age = age
        self.origin = origin
        self.headers = headers
        self.methods = methods

    def on_request(self, context):
        response.headers["Access-Control-Allow-Origin"] = self.origin
        response.headers["Access-Control-Max-Age"] = self.age
        response.headers["Access-Control-Allow-Headers"] = self.headers
        response.headers["Access-Control-Allow-Methods"] = self.methods
        response.headers["Access-Control-Allow-Credentials"] = "true"
        if request.method == "OPTIONS":
            raise HTTP(200)



from ombott.server_adapters import ServerAdapter
from rocket3 import Rocket3

from .logging_utils import logging_conf


class Rocket3Adapter(ServerAdapter):
    "Class implementing a rocket3 server"

    def run(self, handler):
        "runs the server"

        if not self.quiet:
            logging_conf(
                self.options["logging_level"],
            )

        keyfile = self.options.get("keyfile")
        certfile = self.options.get("certfile")

        if keyfile and certfile:
            interface = (self.host, self.port, keyfile, certfile)
        else:
            interface = (self.host, self.port)

        server = Rocket3(interface, "wsgi", dict(wsgi_app=handler))
        server.start()



import base64
import hashlib
import json
import os
import time
import uuid

from py4web import HTTP, request
from py4web.core import Fixture, Session


class URLVerifier(Fixture):
    """This class checks for the validity of URL signatures.
    Specifically, an object of this class can be passed as argument
    to action.uses() to check for the validity of signatures, and the
    sign() method can be used to sign a URL.  If an object of this class
    is passed to the URL helper, it can be used to sign a URL."""

    def __init__(self, url_signer):
        super().__init__()
        if url_signer.session is not None:
            self.__prerequisites__ = [url_signer.session]
        self.url_signer = url_signer

    def on_request(self, context):
        """Checks the request's signature"""
        # extra and remove the signature from the query
        signature = request.query.get("_signature")
        if signature is None:
            raise HTTP(403)
        try:
            h = self.url_signer.algo(self.url_signer.get_key())
            signature = request.query["_signature"]
            sig_content = base64.b16decode(signature.encode("utf-8")).decode("utf-8")
            sig_dict = json.loads(sig_content)
            ts = sig_dict["ts"]
            salt = sig_dict["salt"]
            sig = sig_dict["sig"]
            h.update(
                self.url_signer.get_info_to_sign(
                    request.fullpath, request.query, ts, salt
                )
            )
            computed_sig = base64.b16encode(h.digest()).decode("utf-8")
            if sig != computed_sig:
                raise HTTP(403)
            # We remove the signature, not to pollute the request.
            del request.query["_signature"]
            # Checks the expiration time.
            if self.url_signer.lifespan is not None:
                if float(ts) + self.url_signer.lifespan < time.time():
                    raise HTTP(403)
        except Exception as err:
            raise HTTP(403, body=str(err))

    def _decode_ts(self, ts_string):
        """Decodes the timestamp, removing the salt."""
        s = base64.b16encode(ts_string.encode("utf-8")).decode("utf-8")
        return float(s.split(";")[1])


class URLSigner(Fixture):
    RESERVED_VARIABLES = ["_ts", "_signature"]

    def __init__(
        self,
        session=None,
        key=None,
        variables_to_sign=None,
        signing_info=None,
        lifespan=None,
        algo=None,
    ):
        """
        Signer for URLs.
        :param session: Session.  If a session is not specified, it will use a key
            to sign the URLs.
        :param key: key to sign, used if no session is specified.  If neither a
            session nor a key is specified, then Session.SECRET is used to sign.
        :param variables_to_sign: List of variables to be included in the signature.
        :param signing_info: A function that, when called, returns an additional
            string that is passed into the signing algorithm.  One can e.g. include
            the user id among the things that should not change by doing:
            signing_info = lambda : str(self.session.get("user", {}).get("id", ""))
        :param lifespan: Lifespan of the signature, in seconds.
        :param algo: Algorithm to use for signing. Default is hashlib.sha256().

        The usage is as follows, typically.

        # We build a URL signer.
        url_signer = URLSigner(session)

        @action('/somepath')
        @action.uses(url_signer)
        def somepath():
            # This controller signs a URL.
            return dict(signed_url = URL('/anotherpath', signer=url_signer))

        @action('/anotherpath')
        @action.uses(url_signer.verify())
        def anotherpath():
            # The signature has been verified.
            return dict()
        """
        super().__init__()
        self.session = session
        if session is not None:
            # This ensures that the session will be saved with its changes
            # (including the signing key).
            self.__prerequisites__ = [session]
        self.key = key or Session.SECRET
        self.variables_to_sign = variables_to_sign or []
        self.signing_info = signing_info
        self.lifespan = lifespan
        assert "_signature" not in self.variables_to_sign
        self.algo = algo or hashlib.sha256

    def on_request(self, context):
        """Creates the signing key if necessary."""
        if self.session is not None and self.session.get("_signature_key") is None:
            key = str(uuid.uuid1())
            self.session["_signature_key"] = key
            # Note that we CANNOT save the key in the URLsigner object,
            # because there is only one object -- it's not thread-local in
            # the way the session is.  So we cannot cache it for the later
            # call to get_key.

    def get_key(self):
        """Gets the signing key."""
        if self.session is None:
            key = self.key
        else:
            key = self.session.get("_signature_key")
        assert key is not None, "No signing key"
        return key.encode("utf8")

    def get_info_to_sign(self, url, variables, ts, salt):
        """Gathers the information to be signed."""
        # The key consists of the url, and of the URL parameters.
        return json.dumps(
            {
                "url": url,
                "info": self.signing_info() if self.signing_info is not None else "",
                "vars": {v: str(variables.get(v)) for v in self.variables_to_sign},
                "ts": ts,
                "salt:": salt,
            }
        ).encode("utf-8")

    def sign(self, url, variables):
        """Signs the URL"""
        for v in self.RESERVED_VARIABLES:
            assert v not in variables
        h = self.algo(self.get_key())
        ts = "%.3f" % time.time()
        salt = str(uuid.uuid1())
        url_prefix = os.environ.get("PY4WEB_URL_PREFIX", "")
        if url_prefix:
            url = url_prefix + url
        h.update(self.get_info_to_sign(url, variables, ts, salt))
        sig = base64.b16encode(h.digest()).decode("utf-8")
        sig_content = json.dumps(dict(ts=ts, salt=salt, sig=sig))
        signature = base64.b16encode(sig_content.encode("utf-8")).decode("utf-8")
        variables["_signature"] = signature

    def verify(self):
        """returns a fixture that verifies the URL and optionally the query_keys"""
        return URLVerifier(self)



import os
import unittest

from py4web.core import Template

PATH = os.path.join(os.path.dirname(__file__), "templates")


class TemplateTest(unittest.TestCase):
    def test_template(self):
        t = Template("index.html", path=PATH)
        context = dict(output=dict(n=3), template_inject={})
        t.on_success(context)
        output = context["output"]
        self.assertEqual(output, "0,1,2.\n")
